import logging
from datetime import datetime

from .indexing import ResourceIndexRequest, index_resource
from .ingestion import deletion
from .ingestion.ingesting import ingest_resource
from .utils import check_version_is_valid
from .. import common
from ..datastore_utils import get_latest_version

log = logging.getLogger(__name__)


class ResourceImportRequest(object):
    """
    Class representing a request to import new data into a resource.

    We use a class like this for two reasons, firstly to avoid having a long list of
    arguments passed through to queued functions, and secondly because rq by default
    logs the arguments sent to a function and if the records argument is a large list of
    dicts this becomes insane.
    """

    def __init__(self, resource, version, replace, records=None, api_key=None):
        '''
        :param resource: the resource we're going to import (this must be the resource dict)
        :param version: the version of the resource to import
        :param replace: whether to replace the existing data or not
        :param records: a list of dicts to import, or None if the data is coming from URL or file
        '''
        self.resource = resource
        self.version = version
        self.replace = replace
        self.records = records
        self.api_key = api_key
        self.resource_id = self.resource['id']

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        if self.records is not None:
            records = len(self.records)
        else:
            records = 0
        return (
            f'Import on {self.resource_id}, version {self.version}, replace: {self.replace}, '
            f'records: {records}'
        )


def import_resource_data(request):
    """
    Ingests the resource data into mongo and then, if needed, indexes it into
    elasticsearch. If the data argument is None (note, not falsey or an empty list,
    actually None) then the resource's url field is used as the source of the data.

    This function is blocking so it should be called through the background task queue to avoid
    blocking up a CKAN thread.

    :param request: the ResourceImportRequest object describing the resource import we need to do
    """
    # first, double check that the version is valid
    if not check_version_is_valid(request.resource_id, request.version):
        # log and silently skip this import
        log.info(
            f'Skipped importing data for {request.resource_id} at version {request.version} '
            f'as the version is invalid'
        )
        return

    log.info(
        f'Starting data import for {request.resource_id} at version {request.version}'
    )

    # ingest the resource into mongo
    did_ingest = ingest_resource(
        request.version,
        common.CONFIG,
        request.resource,
        request.records,
        request.replace,
        request.api_key,
    )
    if did_ingest:
        # find out what the latest version in the index is
        latest_index_version = get_latest_version(request.resource_id)

        # index the resource from mongo into elasticsearch. This will only index the records that
        # have changed between the latest index version and the newly ingested version
        index_resource(
            ResourceIndexRequest(
                request.resource, latest_index_version, request.version
            )
        )

        log.info(
            f'Ingest and index complete for {request.resource_id} at version '
            f'{request.version}'
        )


class ResourceDeletionRequest:
    """
    Class representing a request to delete all of a resource's data.

    We use a class like this for two reasons, firstly to avoid having a long list of
    arguments passed through to queued functions, and secondly because rq by default
    logs the arguments sent to a function and if the records argument is a large list of
    dicts this becomes insane.
    """

    def __init__(self, resource, version):
        '''
        :param resource: the resource we're going to delete (this must be the resource dict)
        :param version: the version of the resource to delete
        '''
        self.resource = resource
        self.version = version
        self.resource_id = resource['id']

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f'Deletion of {self.resource_id} at version {self.version}'


def delete_resource_data(request):
    """
    Deletes all the resource's data. This involves ingesting a new version where all
    fields in each record are missing and then indexing this new version.

    This function is blocking so it should be called through the background task queue to avoid
    blocking up a CKAN thread.

    :param request: the ResourceDeletionRequest object describing the resource deletion we need to
                    do
    """
    # first, double check that the version is valid
    if not check_version_is_valid(request.resource_id, request.version):
        # log and silently skip this import
        log.info(
            f'Skipped importing data for {request.resource_id} at version {request.version} '
            f'as the version is invalid'
        )
        return

    log.info(
        f'Starting data deletion for {request.resource_id} at version {request.version}'
    )

    # store a start time, this will be used as the ingestion time of the records
    start = datetime.now()
    # delete the data in mongo
    did_delete = deletion.delete_resource_data(
        request.resource_id, request.version, start
    )
    if did_delete:
        # find out what the latest version in the index is
        latest_index_version = get_latest_version(request.resource_id)

        # index the resource from mongo into elasticsearch. This will only index the records that
        # have changed between the latest index version and the newly ingested version
        index_resource(
            ResourceIndexRequest(
                request.resource, latest_index_version, request.version
            )
        )

        log.info(
            f'Deletion complete for {request.resource_id} at version {request.version}'
        )



import pytest

from ckanext.versioned_datastore.lib.downloads import query
from ckanext.versioned_datastore.logic.actions.meta.arg_objects import QueryArgs
from tests.helpers import patches


class TestDownloadQuery:
    def test_create_query(self):
        with patches.query_schemas():
            q = query.Query(
                query={},
                query_version='v1.0.0',
                resource_ids_and_versions={'resource-id-here': 0},
            )
        assert q.query == {}
        assert q.query_version == 'v1.0.0'
        assert q.resource_ids_and_versions == {'resource-id-here': 0}

    @pytest.mark.ckan_config('ckan.plugins', 'versioned_datastore')
    @pytest.mark.usefixtures('with_plugins', 'with_versioned_datastore_tables')
    def test_create_query_from_query_args(self):
        resource_ids = ['test-resource-id']
        with patches.get_available_resources(
            resource_ids
        ), patches.query_schemas(), patches.rounded_versions():
            query_args = QueryArgs(
                query={
                    'filters': {
                        'and': [
                            {
                                'string_equals': {
                                    'fields': ['collectionCode'],
                                    'value': 'bot',
                                }
                            }
                        ]
                    },
                    'resource_ids': resource_ids,
                }
            )
            q = query.Query.from_query_args(query_args)

        assert q.query_version == 'v1.0.0'
        assert len(q.resource_ids_and_versions) == len(resource_ids)
        assert sorted(q.resource_ids_and_versions.keys()) == resource_ids
        assert 'filters' in q.query
        assert q.hash == '83fc3087623dfd7371cf697d1de6c879de3e722a'



from copy import copy
from datetime import datetime

from elasticsearch_dsl import Search, MultiSearch
from splitgill.search import create_version_query, create_index_specific_version_filter
from splitgill.utils import to_timestamp

from ckan import model
from ckan.plugins import toolkit
from .. import common
from ..datastore_utils import prefix_resource, get_last_after, trim_index_name


def get_available_datastore_resources(context, only=None):
    """
    Returns a set of resource ids accessible to the current user based on the given
    context that are also datastore resources. If the only parameter is passed then it
    is used to filter the set of resources that are returned to include only ones in the
    only list. If the parameter is not passed, or indeed is falsey in any way (such as
    an empty list) then all resource ids available to the user are returned.

    :param context: the dict ckan context to request auth against
    :param only: optional list of resource ids to filter the returned list by. Defaults to None
                 which indicates all available resources should be returned
    :return: a set of resource ids
    """
    # retrieve all resource ids and associated package ids direct from the database for speed
    query = (
        model.Session.query(model.Resource)
        .join(model.Package)
        .filter(model.Resource.state == 'active')
        .filter(model.Package.state == 'active')
        .with_entities(model.Resource.id, model.Package.id)
    )
    # retrieve the names in the status index
    status_search = Search(
        index=common.CONFIG.elasticsearch_status_index_name, using=common.ES_CLIENT
    ).source(['name'])

    if only:
        # apply filters to only get the resources passed in the only list
        query = query.filter(model.Resource.id.in_(only))
        status_search = status_search.filter('terms', name=only)

    # complete the database query and the elasticsearch query
    resources_and_packages = list(query)
    datastore_resources = {hit.name for hit in status_search.scan()}

    # this is the set of resource ids we will populate and return
    resource_ids = set()

    # when a resource is checked in ckan to see if can be accessed by the current user, it's package
    # is inspected. To avoid inspecting the same package over and over again when multiple resources
    # exist under a package, we'll cache the results
    package_access_cache = {}

    for resource_id, package_id in resources_and_packages:
        # if the resource isn't a datastore resource or we've already handled it, ignore it
        if resource_id not in datastore_resources or resource_id in resource_ids:
            continue

        # check the cache, we'll get back either True, False or None
        has_access = package_access_cache.get(package_id, None)

        if has_access:
            # access allowed, add to the list
            resource_ids.add(resource_id)
        elif has_access is None:
            # if the result of looking in the cache is None then there is no value for this package
            # in the cache and we need to do the work
            try:
                try:
                    # remove any cached package from the context so that we know we're
                    # definitely getting the permissions for *this* package
                    del context['package']
                except KeyError:
                    pass
                toolkit.check_access('package_show', context, {'id': package_id})
                package_access_cache[package_id] = True
                # access allowed, add to the list
                resource_ids.add(resource_id)
            except toolkit.NotAuthorized:
                package_access_cache[package_id] = False
        else:
            # skip, there is no access
            continue

    return resource_ids


def determine_resources_to_search(
    context, resource_ids=None, resource_ids_and_versions=None
):
    """
    Determines the resources to search from the given parameters. The set of resource
    ids returned contains only the resources that the user has access to (this is
    determined using the context) and are datastore active. If resource ids are provided
    through either the resource_ids or resource_ids_and_versions parameters then only
    these resource ids will be returned, if indeed they are accessible to the user.

    :param context: the context dict allowing us to determine the user and do auth on the resources
    :param resource_ids: a list of resources to search
    :param resource_ids_and_versions: a dict of resources and versions to search at
    :return: 2-tuple containing a list of resource ids to search and a list of resource ids that
             have been skipped because the user doesn't have access to them or they aren't datastore
             resources
    """
    # validate the resource_ids passed in. If the resource_ids_and_versions parameter is in use then
    # it is taken as the resource_ids source and resource_ids is ignored
    if resource_ids_and_versions:
        requested_resource_ids = list(resource_ids_and_versions.keys())
    else:
        requested_resource_ids = resource_ids
    # this will return the subset of the requested resource ids that the user can search over
    resource_ids = get_available_datastore_resources(context, requested_resource_ids)

    if requested_resource_ids is not None:
        skipped_resources = [
            rid for rid in requested_resource_ids if rid not in resource_ids
        ]
    else:
        skipped_resources = []
    return list(resource_ids), skipped_resources


def determine_version_filter(
    version=None, resource_ids=None, resource_ids_and_versions=None
):
    """
    Determine and return the elasticsearch-dsl filter which can filter on the version
    extracted from the given parameters.

    :param version: the version to filter on across all resources
    :param resource_ids: the resource to search
    :param resource_ids_and_versions: a dict of resource ids -> versions providing resource specific
                                      versions for search
    :return: an elasticsearch-dsl object
    """
    if not resource_ids_and_versions:
        # default the version to now if necessary
        if version is None:
            version = to_timestamp(datetime.now())
        # just use a single version filter if we don't have any resource specific versions
        return create_version_query(version)
    else:
        # run through the resource specific versions provided and ensure they're rounded down
        indexes_and_versions = {}
        for resource_id in resource_ids:
            target_version = resource_ids_and_versions[resource_id]
            if target_version is None:
                raise toolkit.ValidationError(
                    f'Valid version not given for {resource_id}'
                )
            index = prefix_resource(resource_id)
            rounded_version = common.SEARCH_HELPER.get_rounded_versions(
                [index], target_version
            )[index]
            indexes_and_versions[index] = rounded_version

        return create_index_specific_version_filter(indexes_and_versions)


def calculate_after(result, size):
    """
    Calculate the after value for the given search result. It is assumed that the size
    used when the search was completed is 1 larger than the size passed as a parameter
    to this function.

    :param result: the elasticsearch result object
    :param size: the number of results
    :return: a 2-tuple containing the list of hits and the next after value
    """
    if len(result.hits) > size:
        # there are more results, trim off the last hit as it wasn't requested
        hits = result.hits[:-1]
        next_after = get_last_after(hits)
    else:
        # there are no more results beyond the ones we're going to pass back
        next_after = None
        hits = result.hits
    return hits, next_after


def chunk_iterator(iterable, chunk_size):
    """
    Iterates over an iterable, yielding lists of size chunk_size until the iterable is
    exhausted. The final list could be smaller than chunk_size but will always have a
    length > 0.

    :param iterable: the iterable to chunk up
    :param chunk_size: the maximum size of each yielded chunk
    :return: a generator of list chunks
    """
    chunk = []
    for element in iterable:
        chunk.append(element)
        if len(chunk) == chunk_size:
            yield chunk
            chunk = []
    if chunk:
        yield chunk


def find_searched_resources(search, resource_ids):
    """
    Given a search and a list of resource ids to search in, returns a list of the
    resources that are actually included in the search results.

    :param search: an elasticsearch-dsl object
    :param resource_ids: a list of resource ids
    :return: a list of resource ids
    """
    # we have to make a copy as aggs don't return a clone :(
    search_copy = copy(search)
    search_copy = search_copy.index(
        [prefix_resource(resource_id) for resource_id in resource_ids]
    )
    search_copy.aggs.bucket('indexes', 'terms', field='_index')
    multisearch = MultiSearch(using=common.ES_CLIENT).add(search_copy)
    result = next(iter(multisearch.execute()))
    return [
        trim_index_name(bucket['key'])
        for bucket in result.aggs.to_dict()['indexes']['buckets']
        if bucket['doc_count'] > 0
    ]


def get_resources_and_versions(
    resource_ids=None,
    resource_ids_and_versions=None,
    version=None,
    allow_non_datastore=False,
):
    """
    Get a list of resource ids and a dict of resource ids and versions from either, e.g.
    get the list of resource ids from a resource id and version dict.

    :param resource_ids: a list of resource ids
    :param resource_ids_and_versions: a dict of resource id: resource version
    :param version: a datestamp used as a default version for resources without a version
    :param allow_non_datastore: allow non datastore resources to be included (will be
                                returned with common.NON_DATASTORE_VERSION)
    :return: a tuple of resource_ids, resource_ids_and_versions
    """

    if resource_ids_and_versions is None:
        resource_ids_and_versions = {}
    else:
        # use the resource_ids_and_versions dict first over the resource_ids and version params
        resource_ids = list(resource_ids_and_versions.keys())

    # first see what's available from the datastore
    available_resource_ids = list(get_available_datastore_resources({}, resource_ids))
    if (not available_resource_ids) and (not allow_non_datastore):
        raise toolkit.ValidationError(
            "The requested resources aren't accessible to this user"
        )

    unavailable_resource_ids = [
        rid for rid in resource_ids or [] if rid not in available_resource_ids
    ]
    non_datastore_resources = []

    if allow_non_datastore:
        resource_show = toolkit.get_action('resource_show')
        for resource_id in unavailable_resource_ids:
            resource = resource_show({}, {'id': resource_id})
            # if we get nothing back there's probably an access issue; if it's a
            # datastore resource something went wrong earlier
            if resource and not resource['datastore_active']:
                available_resource_ids.append(resource_id)
                non_datastore_resources.append(resource_id)

    rounded_resource_ids_and_versions = {}
    # see if a version was provided; we'll use this if a resource id we're searching doesn't
    # have a directly assigned version (i.e. it was absent from the resource_ids_and_versions
    # dict, or that parameter wasn't provided)
    if version is None:
        version = to_timestamp(datetime.now())
    for resource_id in available_resource_ids:
        if resource_id in non_datastore_resources:
            rounded_resource_ids_and_versions[
                resource_id
            ] = common.NON_DATASTORE_VERSION
            continue
        # try to get the target version from the passed resource_ids_and_versions dict, but if
        # it's not in there, default to the version variable
        target_version = resource_ids_and_versions.get(resource_id, version)
        index = prefix_resource(resource_id)
        # round the version down to ensure we search the exact version requested
        rounded_version = common.SEARCH_HELPER.get_rounded_versions(
            [index], target_version
        )[index]
        if rounded_version is not None:
            # resource ids without a rounded version are skipped
            rounded_resource_ids_and_versions[resource_id] = rounded_version

    return available_resource_ids, rounded_resource_ids_and_versions



import csv

from .base import BaseDerivativeGenerator
from ..utils import flatten_dict


class CsvDerivativeGenerator(BaseDerivativeGenerator):
    name = 'csv'
    extension = 'csv'

    def __init__(
        self,
        output_dir,
        fields,
        query,
        resource_id=None,
        delimiter='comma',
        **format_args
    ):
        super(CsvDerivativeGenerator, self).__init__(
            output_dir, fields, query, resource_id, delimiter='comma', **format_args
        )
        self.delimiter = {'comma': ',', 'tab': '\t'}[delimiter]
        self.writer = None

    def initialise(self):
        self.writer.writeheader()
        super(CsvDerivativeGenerator, self).initialise()

    def setup(self):
        self.writer = csv.DictWriter(
            self.main_file, self.fields['main'], delimiter=self.delimiter
        )
        super(CsvDerivativeGenerator, self).setup()

    def finalise(self):
        self.writer = None
        super(CsvDerivativeGenerator, self).finalise()

    def _write(self, record):
        row = flatten_dict(record)
        filtered_row = {}
        for field, value in row.items():
            if value is None and field not in self.fields['main']:
                continue
            elif field not in self.fields['main']:
                raise ValueError('Unexpected field.')
            else:
                filtered_row[field] = value
        if self.resource_id:
            filtered_row[self.RESOURCE_ID_FIELD_NAME] = self.resource_id
        self.writer.writerow(filtered_row)



import itertools
from collections import namedtuple
from datetime import datetime as dt
from threading import Thread

next_id = itertools.count(start=1)
Job = namedtuple('Job', ['enqueued_at', 'id'])


def sync_enqueue_job(job_func, args=None, kwargs=None, *queue_args, **queue_kwargs):
    args = args or []
    kwargs = kwargs or {}
    job_func(*args, **kwargs)
    return Job(enqueued_at=dt.now(), id=1)


def sync_enqueue_job_thread(
    job_func, args=None, kwargs=None, *queue_args, **queue_kwargs
):
    args = args or []
    kwargs = kwargs or {}

    job_thread = Thread(target=job_func, args=args, kwargs=kwargs)
    job_thread.start()
    job_thread.join()

    return Job(enqueued_at=dt.now(), id=next(next_id))



from mock import MagicMock

from ckanext.versioned_datastore.lib.importing.ingestion.records import DatastoreRecord


def test_convert_empty():
    scenarios = [
        ({}, {}),
        (
            {'field1': 'beans', 'field2': [1, 2, 3, 4]},
            {'field1': 'beans', 'field2': [1, 2, 3, 4]},
        ),
        ({'field1.another': 'beans'}, {'field1_another': 'beans'}),
        (
            {'field1.another.more.moreagain': 'beans'},
            {'field1_another_more_moreagain': 'beans'},
        ),
        ({'': 'beans'}, {}),
        (
            {'field': {'field1': 'beans', 'field2.field3': False, 'a': {'': 'beans'}}},
            {'field': {'field1': 'beans', 'field2_field3': False, 'a': {}}},
        ),
    ]

    for data, expected_result in scenarios:
        record = DatastoreRecord(MagicMock(), MagicMock(), data, MagicMock())
        converted = record.convert()
        assert converted == expected_result


def test_simple_things():
    record_id = MagicMock()
    resource_id = MagicMock()
    record = DatastoreRecord(MagicMock(), record_id, MagicMock(), resource_id)
    assert record.id == record_id
    assert record.mongo_collection == resource_id



import hashlib
import io
import json
import string
from collections import defaultdict

from elasticsearch_dsl import Search
from elasticsearch_dsl.query import Bool, Q

from .schema import Schema, load_core_schema, schema_base_path
from ..datastore_utils import prefix_field


class v1_0_0Schema(Schema):
    """
    Schema class for the v1.0.0 query schema.
    """

    version = 'v1.0.0'

    def __init__(self):
        self.schema, self.validator = load_core_schema(v1_0_0Schema.version)
        self.geojson = {
            'country': self.load_geojson(
                '50m-admin-0-countries-v4.1.0.geojson', ('NAME_EN', 'NAME')
            ),
            # if we use name_en we end up with one atlantic ocean whereas if we use name we get 2 -
            # the "North Atlantic Ocean" and the "South Atlantic Ocean". I think this is preferable.
            'marine': self.load_geojson('50m-marine-regions-v4.1.0.geojson', ('name',)),
            'geography': self.load_geojson(
                '50m-geography-regions-v4.1.0.geojson', ('name_en', 'name')
            ),
        }
        self.hasher = v1_0_0Hasher()

    def validate(self, query):
        """
        Validates the query against the v1.0.0 schema.

        :param query: the query to validate
        """
        self.validator.validate(query)

    def hash(self, query):
        """
        Hashes the given query and returns the hex digest of it.

        :param query: the query dict
        :return: the hex digest of the hash of the query
        """
        return self.hasher.hash_query(query)

    def translate(self, query, search=None):
        """
        Translates the query into an elasticsearch-dsl search object.

        :param query: the whole query dict
        :param search: an instantiated elasticsearch-dsl object to be built on instead of creating
                       a fresh object. By default a new search object is created.
        :return: an instantiated elasticsearch-dsl object
        """
        search = Search() if search is None else search
        search = self.add_search(query, search)
        search = self.add_filters(query, search)
        return search

    def add_search(self, query, search):
        """
        Adds a search to the search object and then returns it. Search terms map
        directly to the elasticsearch match query on the meta.all field. If there is no
        search in the query then the search object passed in is simply returned
        unaltered.

        :param query: the whole query dict
        :param search: an instantiated elasticsearch-dsl object
        :return: an instantiated elasticsearch-dsl object
        """
        if 'search' in query:
            return search.query(
                'match', **{'meta.all': {'query': query['search'], 'operator': 'and'}}
            )
        return search

    def add_filters(self, query, search):
        """
        Adds filters from the query into the search object and then returns it. If no
        filters are defined in the query then the search object passed in is simply
        returned unaltered.

        :param query: the whole query dict
        :param search: an instantiated elasticsearch-dsl object
        :return: an instantiated elasticsearch-dsl object
        """
        if 'filters' in query:
            return search.query(self.create_group_or_term(query['filters']))
        return search

    def create_group_or_term(self, group_or_term):
        """
        Creates and returns the elasticsearch-dsl query object necessary for the given
        group or term dict and returns it.

        :param group_or_term: a dict defining a single group or term
        :return: an elasticsearch-dsl object such as a Bool or Query object
        """
        # only one property is allowed so we can safely just extract the only name and options
        group_or_term_type, group_or_term_options = next(iter(group_or_term.items()))
        return getattr(self, f'create_{group_or_term_type}')(group_or_term_options)

    def create_and(self, group):
        """
        Creates and returns an elasticsearch-dsl query object representing the given
        group as an and query. This will be a Bool with a must in it for groups with
        more than 1 member, or will just be the actual member if only 1 member is found
        in the group. This is strictly unnecessary as elasticsearch/lucerne itself will
        normalise the query and remove redundant nestings but we might as well do it
        here seeing as we can and it makes smaller elasticsearch queries.

        :param group: the group to build the and from
        :return: the first member from the group if there's only one member in the group, or a Bool
        """
        members = [self.create_group_or_term(member) for member in group]
        return members[0] if len(members) == 1 else Bool(filter=members)

    def create_or(self, group):
        """
        Creates and returns an elasticsearch-dsl query object representing the given
        group as an or query. This will be a Bool with a should in it for groups with
        more than 1 member, or will just be the actual member if only 1 member is found
        in the group. This is strictly unnecessary as elasticsearch/lucerne itself will
        normalise the query and remove redundant nestings but we might as well do it
        here seeing as we can and it makes smaller elasticsearch queries.

        :param group: the group to build the or from
        :return: the first member from the group if there's only one member in the group, or a Bool
        """
        return self.build_or([self.create_group_or_term(member) for member in group])

    def create_not(self, group):
        """
        Creates and returns an elasticsearch-dsl query object representing the given
        group as a not query. This will be a Bool with a must_not in it.

        :param group: the group to build the not from
        :return: a Bool query
        """
        return Bool(must_not=[self.create_group_or_term(member) for member in group])

    def create_string_equals(self, options):
        """
        Given the options for a string_equals term, creates and returns an
        elasticsearch-dsl object to represent it. This term maps directly to an
        elasticsearch term query. If only one field is present in the fields property
        then the term query is returned directly, otherwise an or query is returned
        across all the fields requested.

        :param options: the options for the string_equals query
        :return: an elasticsearch-dsl Query object or a Bool object
        """
        return self.build_or(
            [
                Q('term', **{prefix_field(field): options['value']})
                for field in options['fields']
            ]
        )

    def create_string_contains(self, options):
        """
        Given the options for a string_contains term, creates and returns an
        elasticsearch-dsl object to represent it. This term maps directly to an
        elasticsearch match query on the .full subfield. If only one field is present in
        the fields property then the term query is returned directly, otherwise an or
        query is returned across all the fields requested.

        :param options: the options for the string_contains query
        :return: an elasticsearch-dsl Query object or a Bool object
        """
        fields = options['fields']
        query = {'query': options['value'], 'operator': 'and'}

        if fields:
            return self.build_or(
                [
                    Q('match', **{f'{prefix_field(field)}.full': query})
                    for field in fields
                ]
            )
        else:
            return Q('match', **{'meta.all': query})

    def create_number_equals(self, options):
        """
        Given the options for a number_equals term, creates and returns an
        elasticsearch-dsl object to represent it. This term maps directly to an
        elasticsearch term query. If only one field is present in the fields property
        then the term query is returned directly, otherwise an or query is returned
        across all the fields requested.

        :param options: the options for the number_equals query
        :return: an elasticsearch-dsl Query object or a Bool object
        """
        return self.build_or(
            [
                Q('term', **{f'{prefix_field(field)}.number': options['value']})
                for field in options['fields']
            ]
        )

    def create_number_range(self, options):
        """
        Given the options for a number_range term, creates and returns an elasticsearch-
        dsl object to represent it. This term maps directly to an elasticsearch range
        query. If only one field is present in the fields property then the term query
        is returned directly, otherwise an or query is returned across all the fields
        requested.

        :param options: the options for the number_range query
        :return: an elasticsearch-dsl Query object or a Bool object
        """
        less_than = options.get('less_than', None)
        greater_than = options.get('greater_than', None)
        less_than_inclusive = options.get('less_than_inclusive', True)
        greater_than_inclusive = options.get('greater_than_inclusive', True)
        query = {}
        if less_than is not None:
            query['lt' if not less_than_inclusive else 'lte'] = less_than
        if greater_than is not None:
            query['gt' if not greater_than_inclusive else 'gte'] = greater_than

        return self.build_or(
            [
                Q('range', **{f'{prefix_field(field)}.number': query})
                for field in options['fields']
            ]
        )

    def create_exists(self, options):
        """
        Given the options for an exists term, creates and returns an elasticsearch-dsl
        object to represent it. This term maps directly to an elasticsearch exists
        query. If only one field is present in the fields property then the term query
        is returned directly, otherwise an or query is returned across all the fields
        requested.

        :param options: the options for the exists query
        :return: an elasticsearch-dsl Query object or a Bool object
        """
        # TODO: should we provide exists on subfields?
        if options.get('geo_field', False):
            return Q('exists', field='meta.geo')
        else:
            return self.build_or(
                [Q('exists', field=prefix_field(field)) for field in options['fields']]
            )

    def create_geo_point(self, options):
        """
        Given the options for an geo_point term, creates and returns an elasticsearch-
        dsl object to represent it. This term maps directly to an elasticsearch
        geo_distance query. If only one field is present in the fields property then the
        term query is returned directly, otherwise an or query is returned across all
        the fields requested.

        :param options: the options for the geo_point query
        :return: an elasticsearch-dsl Query object or a Bool object
        """
        return Q(
            'geo_distance',
            **{
                'distance': f'{options.get("radius", 0)}{options.get("radius_unit", "m")}',
                'meta.geo': {
                    'lat': options['latitude'],
                    'lon': options['longitude'],
                },
            },
        )

    def create_geo_named_area(self, options):
        """
        Given the options for a geo_named_area term, creates and returns an
        elasticsearch-dsl object to represent it. This term maps directly to one or more
        elasticsearch geo_polygon queries, if necessary combined using ands, ors and
        nots to provide MultiPolygon hole support.

        In v1.0.0, Natural Earth Data datasets are used to provide the lists of names and
        corresponding geojson areas. The 1:50million scale is used in an attempt to provide a good
        level of detail without destroying Elasticsearch with enormous numbers of points. See the
        `theme/public/querySchemas/geojson/` directory for source data and readme, and also the
        load_geojson function in this class.

        :param options: the options for the geo_named_area query
        :return: an elasticsearch-dsl Query object (a single geo_polygon Query or a Bool Query)
        """
        category, name = next(iter(options.items()))
        return self.build_multipolygon_query(self.geojson[category][name])

    def create_geo_custom_area(self, coordinates):
        """
        Given the coordinates for a geo_custom_area term, creates and returns an
        elasticsearch-dsl object to represent it. This term takes the equivalent of the
        coordinates array from a MultiPolygon type feature in GeoJSON and uses it to
        build a query which captures records that fall in the polygon (and outside any
        holes defined in the Polygon).

        :param coordinates: a MultiPolygon coordinates list
        :return: an elasticsearch-dsl Query object (a single geo_polygon Query or a Bool Query)
        """
        return self.build_multipolygon_query(coordinates)

    @staticmethod
    def build_or(terms):
        """
        Utility function which when given a list of elasticsearch-dsl query objects,
        either returns the first one on it's own or creates an "or" query encapsulating
        them.

        :param terms: a list of elasticsearch-dsl terms
        :return: either a Query object or a Bool should object
        """
        return (
            terms[0] if len(terms) == 1 else Bool(should=terms, minimum_should_match=1)
        )

    @staticmethod
    def build_geo_polygon_query(points):
        """
        Given a list of points (where each point is a list with 2 elements, the
        longitude and the latitude (note the order, it's the same as the GeoJSON spec)),
        creates a geo_polygon elasticsearch-dsl query object for the points and returns
        it.

        :param points: a list of points
        :return: an elasticsearch-dsl query object
        """
        return Q(
            'geo_polygon',
            **{
                'meta.geo': {
                    'points': [{'lat': point[1], 'lon': point[0]} for point in points]
                }
            },
        )

    @staticmethod
    def build_multipolygon_query(coordinates):
        """
        Utility function for building elasticsearch-dsl queries that represent GeoJSON
        MultiPolygons. Given the coordinates this function creates a geo_polygon queries
        and Bool queries to represent the varioud enclosures and holes in those
        enclosures to find all records residing in the MultiPolygon. The coordinates
        parameter should match the format required by GeoJSON and therefore be a series
        of nested lists, see the GeoJSON docs for details.

        :param coordinates: the coordinate list, which is basically a list of Polygons. See the
                            GeoJSON doc for the exact format and meaning
        :return: an elasticsearch-dsl object representing the MultiPolygon
        """
        queries = []
        # the first list is a list of GeoJSON Polygons
        for polygon in coordinates:
            # then the Polygon is a list containing at least one element. The first element is the
            # outer boundary shape of the polygon and any other elements are holes in this shape
            outer, holes = polygon[0], polygon[1:]
            outer_query = v1_0_0Schema.build_geo_polygon_query(outer)

            if holes:
                holes_queries = [
                    v1_0_0Schema.build_geo_polygon_query(hole) for hole in holes
                ]
                # create a query which filters the outer query but filters out the holes
                queries.append(Bool(filter=[outer_query], must_not=holes_queries))
            else:
                queries.append(outer_query)

        return v1_0_0Schema.build_or(queries)

    @staticmethod
    def load_geojson(filename, name_keys):
        """
        Load the given geojson file, build a lookup using the data and the name_keys
        parameter and return it.

        The geojson file is assumed to be a list of features containing only Polygon or
        MultiPolygon types.

        The name_keys parameter should be a sequence of keys to use to retrieve a name for the
        feature from the properties dict. The first key found in the properties dict with a value is
        used and therefore the keys listed should be in priority order. The extracted name is passed
        to string.capwords to produce a sensible and consistent set of names.

        :param filename: the name geojson file to load from the given path
        :param name_keys: a priority ordered sequence of keys to use for feature name retrieval
        :return: a dict of names -> MultiPolygons
        """
        path = schema_base_path.joinpath(v1_0_0Schema.version).joinpath('geojson')

        # make sure we read the file using utf-8
        with io.open(path.joinpath(filename), 'r', encoding='utf-8') as f:
            lookup = defaultdict(list)
            for feature in json.load(f)['features']:
                # find the first name key with a value and pass it to string.capwords
                name = string.capwords(
                    next(
                        iter(
                            filter(
                                None,
                                (
                                    feature['properties'].get(key, None)
                                    for key in name_keys
                                ),
                            )
                        )
                    )
                )

                coordinates = feature['geometry']['coordinates']
                # if the feature is a Polygon, wrap it in a list to make it a MultiPolygon
                if feature['geometry']['type'] == 'Polygon':
                    coordinates = [coordinates]

                # add the polygons found to the existing MultiPolygon (some names are listed
                # multiple times in the source geojson files and require stitching together to make
                # a single name -> MultiPolygon mapping
                for polygon in coordinates:
                    # if a polygon is already represented in the MultiPolygon, ignore the dupe
                    if polygon not in lookup[name]:
                        lookup[name].append(polygon)

            return lookup


class v1_0_0Hasher:
    """
    Query hasher class for the v1.0.0 query schema.
    """

    def hash_query(self, query):
        """
        Stable hash function for v1.0.0 queries.

        :param query: the query dict
        :return: the hex digest
        """
        query_hash = hashlib.sha1()
        if 'search' in query:
            query_hash.update(f'search:{query["search"]}'.encode(u'utf-8'))
        if 'filters' in query:
            data = f'filters:{self.create_group_or_term(query["filters"])}'.encode(
                u'utf-8'
            )
            query_hash.update(data)
        return query_hash.hexdigest()

    def create_group_or_term(self, group_or_term):
        """
        Creates and returns a string version of the given group or term dict and returns
        it.

        :param group_or_term: a dict defining a single group or term
        :return: a string representing the group or term
        """
        # only one property is allowed so we can safely just extract the only name and options
        group_or_term_type, group_or_term_options = next(iter(group_or_term.items()))
        return getattr(self, f'create_{group_or_term_type}')(group_or_term_options)

    def create_and(self, group):
        """
        Creates and returns a string version of the given group as an and query.

        :param group: the group to build the and from
        :return: a string representing the group
        """
        # sorting the members makes this stable
        members = sorted(self.create_group_or_term(member) for member in group)
        return f'and:[{"|".join(members)}]'

    def create_or(self, group):
        """
        Creates and returns a string version of the given group as an or query.

        :param group: the group to build the or from
        :return: a string representing the group
        """
        # sorting the members makes this stable
        members = sorted(self.create_group_or_term(member) for member in group)
        return f'or:[{"|".join(members)}]'

    def create_not(self, group):
        """
        Creates and returns a string version of the given group as a not query.

        :param group: the group to build the not from
        :return: a string representing the group
        """
        # sorting the members makes this stable
        members = sorted(self.create_group_or_term(member) for member in group)
        return f'not:[{"|".join(members)}]'

    @staticmethod
    def create_string_equals(options):
        """
        Given the options for a string_equals term, creates and returns a string version
        of it.

        :param options: the options for the string_equals query
        :return: a string representing the term
        """
        # sorting the fields makes this stable
        fields = ','.join(sorted(options['fields']))
        return f'string_equals:{fields};{options["value"]}'

    @staticmethod
    def create_string_contains(options):
        """
        Given the options for a string_contains term, creates and returns a string
        version of it.

        :param options: the options for the string_contains query
        :return: a string representing the term
        """
        # sorting the fields makes this stable
        fields = ','.join(sorted(options['fields']))
        return f'string_contains:{fields};{options["value"]}'

    @staticmethod
    def create_number_equals(options):
        """
        Given the options for a number_equals term, creates and returns a string version
        of it.

        :param options: the options for the number_equals query
        :return: a string representing the term
        """
        # sorting the fields makes this stable
        fields = ','.join(sorted(options['fields']))
        return f'number_equals:{fields};{options["value"]}'

    @staticmethod
    def create_number_range(options):
        """
        Given the options for a number_range term, creates and returns a string version
        of it.

        :param options: the options for the number_range query
        :return: a string representing the term
        """
        # sorting the fields makes this stable
        fields = ','.join(sorted(options['fields']))
        hash_value = f'number_range:{fields};'

        less_than = options.get('less_than', None)
        less_than_inclusive = options.get('less_than_inclusive', True)
        if less_than is not None:
            hash_value += '<'
            if less_than_inclusive:
                hash_value += '='
            hash_value += str(less_than)

        greater_than = options.get('greater_than', None)
        greater_than_inclusive = options.get('greater_than_inclusive', True)
        if greater_than is not None:
            hash_value += '>'
            if greater_than_inclusive:
                hash_value += '='
            hash_value += str(greater_than)

        return hash_value

    @staticmethod
    def create_exists(options):
        """
        Given the options for a exists term, creates and returns a string version of it.

        :param options: the options for the exists query
        :return: a string representing the term
        """
        if options.get('geo_field', False):
            return 'geo_exists'
        else:
            # sorting the fields makes this stable
            fields = ','.join(sorted(options['fields']))
            return f'exists:{fields}'

    @staticmethod
    def create_geo_point(options):
        """
        Given the options for a geo_point term, creates and returns a string version of
        it.

        :param options: the options for the geo_point query
        :return: a string representing the term
        """
        distance = f'{options.get("radius", 0)}{options.get("radius_unit", "m")}'
        return f'geo_point:{distance};{options["latitude"]};{options["longitude"]}'

    @staticmethod
    def create_geo_named_area(options):
        """
        Given the options for a geo_named_area term, creates and returns a string
        version of it.

        :param options: the options for the geo_named_area query
        :return: a string representing the term
        """
        return 'geo_named_area:{};{}'.format(*next(iter(options.items())))

    def create_geo_custom_area(self, coordinates):
        """
        Given the coordinates for a geo_custom_area term, creates and returns a string
        version of it.

        :param coordinates: the coordinates for the geo_custom_area query
        :return: a string representing the term
        """
        queries = []
        # the first list is a list of GeoJSON Polygons
        for polygon in coordinates:
            # then the Polygon is a list containing at least one element. The first element is the
            # outer boundary shape of the polygon and any other elements are holes in this shape
            outer, holes = polygon[0], polygon[1:]
            outer_query = self.build_geo_polygon_query(outer)

            if holes:
                # sort the holes to ensure stability
                holes_queries = sorted(
                    self.build_geo_polygon_query(hole) for hole in holes
                )
                # create a query which filters the outer query but filters out the holes
                queries.append(f'{outer_query}/{holes_queries}')
            else:
                queries.append(outer_query)

        return f'geo_custom_area:{";".join(queries)}'

    @staticmethod
    def build_geo_polygon_query(points):
        """
        Given a series of points, returns a string version of them. Note that we don't
        sort them as that could change the meaning.

        :param points: the points as lat lon pairs
        :return: a string representing the points
        """
        return ','.join(f'[{point[1]},{point[0]}]' for point in points)



import pytest
from ckan.plugins import toolkit
from ckan.tests import factories
from mock import MagicMock, patch

from ckanext.versioned_datastore.lib.downloads import utils, query
from tests.helpers import patches


class TestGetSchema:
    @pytest.mark.ckan_config('ckan.plugins', 'versioned_datastore')
    @pytest.mark.usefixtures('with_plugins', 'clean_db')
    def test_get_schema(self):
        resource_dict = factories.Resource()
        with patches.query_schemas():
            q = query.Query(
                query={
                    'filters': {
                        'and': [
                            {
                                'string_equals': {
                                    'fields': ['collectionCode'],
                                    'value': 'bot',
                                }
                            }
                        ]
                    }
                },
                query_version='v1.0.0',
                resource_ids_and_versions={resource_dict['id']: 1},
            )

        index_name = (
            toolkit.config.get('ckanext.versioned_datastore.elasticsearch_index_prefix')
            + resource_dict['id']
        )

        # this is a _very_ stripped down version of the return value from
        # indices.get_mapping()
        get_mapping_mock = MagicMock(
            return_value={
                index_name: {
                    'mappings': {
                        '_doc': {
                            'properties': {
                                'data': {
                                    'properties': {
                                        '_id': {'type': 'long'},
                                        'name': {
                                            'type': 'keyword',
                                            'fields': {
                                                'full': {'type': 'text'},
                                                'number': {
                                                    'type': 'double',
                                                    'ignore_malformed': True,
                                                },
                                            },
                                            'copy_to': ['meta.all'],
                                            'ignore_above': 256,
                                            'normalizer': 'lowercase_normalizer',
                                        },
                                        'modified': {
                                            'type': 'date',
                                            'format': 'epoch_millis',
                                        },
                                    }
                                }
                            }
                        }
                    }
                }
            }
        )
        with patch(
            "ckanext.versioned_datastore.lib.downloads.utils.get_mappings",
            get_mapping_mock,
        ):
            parsed_schemas = utils.get_schemas(q)
        parsed_schema = parsed_schemas[resource_dict["id"]]
        assert isinstance(parsed_schema, dict)
        assert parsed_schema['type'] == 'record'
        assert parsed_schema['name'] == 'ResourceRecord'
        assert len(parsed_schema['fields']) == 3
        assert isinstance(parsed_schema['fields'][0]['type'], list)


class TestFilterDataFields:
    def test_excludes_field_with_zero_count(self):
        data = {'one': 'a', 'two': 'b', 'three': 'c'}
        field_counts = {'one': 1, 'two': 1, 'three': 0}

        filtered_data = utils.filter_data_fields(data, field_counts)
        assert 'one' in filtered_data
        assert 'two' in filtered_data
        assert 'three' not in filtered_data

    def test_excludes_field_with_no_count(self):
        data = {'one': 'a', 'two': 'b', 'three': 'c'}
        field_counts = {'one': 1, 'two': 1}

        filtered_data = utils.filter_data_fields(data, field_counts)
        assert 'one' in filtered_data
        assert 'two' in filtered_data
        assert 'three' not in filtered_data

    def test_includes_null_field_with_count(self):
        data = {'one': 'a', 'two': 'b', 'three': None}
        field_counts = {'one': 1, 'two': 1, 'three': 1}

        filtered_data = utils.filter_data_fields(data, field_counts)
        assert 'one' in filtered_data
        assert 'two' in filtered_data
        assert 'three' in filtered_data

    def test_excludes_empty_nested_fields(self):
        data = {'one': {'two': 'a', 'three': 'b'}}
        field_counts = {'one.two': 1}

        filtered_data = utils.filter_data_fields(data, field_counts)
        assert 'one' in filtered_data
        assert 'two' in filtered_data['one']
        assert 'three' not in filtered_data['one']

    def test_excludes_empty_listed_fields(self):
        data = {'one': [{'two': 'a', 'three': 'b'}, {'two': 'c', 'four': 'd'}]}
        field_counts = {'one.two': 2, 'one.four': 1}

        filtered_data = utils.filter_data_fields(data, field_counts)
        assert 'one' in filtered_data
        assert len(filtered_data['one']) == 2
        assert 'four' in filtered_data['one'][1]
        for r in filtered_data['one']:
            assert 'two' in r
            assert 'three' not in r


class TestFlattenDict:
    def test_flattens_dict(self):
        data = {'one': {'two': 2, 'three': 3}}
        expected_output = {'one.two': 2, 'one.three': 3}

        flattened = utils.flatten_dict(data)
        assert len(flattened) == len(expected_output)
        for k, v in expected_output.items():
            assert k in flattened
            assert v == flattened[k]

    def test_flattens_nested_lists(self):
        data = {'one': [{'two': 'a', 'three': 'b'}, {'two': 'c', 'four': 'd'}]}
        expected_output = {'one.two': 'a | c', 'one.three': 'b', 'one.four': 'd'}

        flattened = utils.flatten_dict(data)
        assert len(flattened) == len(expected_output)
        for k, v in expected_output.items():
            assert k in flattened
            assert v == flattened[k]

    def test_flattens_lists(self):
        data = {'one': ['a', 'b']}
        expected_output = {'one': 'a | b'}

        flattened = utils.flatten_dict(data)
        assert len(flattened) == len(expected_output)
        for k, v in expected_output.items():
            assert k in flattened
            assert v == flattened[k]



from traceback import format_exception_only

from ckan import model
from datetime import datetime
from sqlalchemy import desc

from ...model.stats import ImportStats

PREP = 'prep'
INGEST = 'ingest'
INDEX = 'index'
ALL_TYPES = [PREP, INDEX, INGEST]


def start_operation(resource_id, import_type, version, start=None):
    """
    Creates an ImportStats instance, saves it to the database and returns the database
    id of the newly created object.

    :param resource_id: the id of the resource being worked on
    :param import_type: the type of import operation being undertaken
    :param version: the version of the data
    :param start: the datetime when this operation was started (optional, if None current time will
                  be used)
    :return: the database id of the saved ImportStats object
    """
    if start is None:
        start = datetime.now()
    stats = ImportStats(
        resource_id=resource_id,
        type=import_type,
        version=version,
        in_progress=True,
        start=start,
    )
    stats.add()
    stats.commit()
    return stats.id


def update_stats(stats_id, update):
    """
    Update the ImportStats object with the given database id with the given update dict.
    The update dict will be passed directly to SQLAlchemy.

    :param stats_id: the database id of the object to update
    :param update: a dict of updates to apply
    """
    model.Session.query(ImportStats).filter(ImportStats.id == stats_id).update(update)
    model.Session.commit()


def finish_operation(stats_id, total, stats=None):
    """
    Update the ImportStats object with the given id to indicate that the operation is
    complete.

    :param stats_id: the database id of the object to finish
    :param total: the total number of records affected by this operation
    :param stats: the stats dict returned by the operation (optional)
    """
    if stats is None:
        start = model.Session.query(ImportStats).get(stats_id).start
        end = datetime.now()
        stats = {
            'duration': (end - start).total_seconds(),
            'start': start,
            'end': end,
            'operations': {},
        }
    update_stats(
        stats_id,
        {
            ImportStats.in_progress: False,
            ImportStats.count: total,
            ImportStats.duration: stats['duration'],
            ImportStats.start: stats['start'],
            ImportStats.end: stats['end'],
            ImportStats.operations: stats['operations'],
        },
    )


def monitor_ingestion(stats_id, ingester):
    """
    Adds monitoring functions to the ingester and updates the ImportStats (with the
    given database id) when updates come through.

    :param stats_id: the database id of the object to update
    :param ingester: the Ingester object to monitor
    """

    @ingester.totals_signal.connect_via(ingester)
    def on_ingest(_sender, total, inserted, updated):
        # this function is called each time a batch of records is ingested from the feeder into
        # mongo. This is done in batches and therefore we don't have to limit the frequency of our
        # database updates
        update_stats(
            stats_id,
            {
                ImportStats.in_progress: True,
                ImportStats.count: total,
            },
        )

    @ingester.finish_signal.connect_via(ingester)
    def on_finish(_sender, total, inserted, updated, stats):
        # this function is called when the ingestion operation completes
        finish_operation(stats_id, total, stats)


def monitor_indexing(stats_id, indexer, update_frequency=1000):
    """
    Adds monitoring functions to the indexer and updates the ImportStats (with the given
    database id) when updates come through.

    :param stats_id: the database id of the object to update
    :param indexer: the Indexer object to monitor
    :param update_frequency: the frequency with which to update the ImportStats. Setting this too
                             low will cause the database written to a lot which could cause
                             performance issues.
    """

    @indexer.index_signal.connect_via(indexer)
    def update_progress(_sender, indexing_stats, **kwargs):
        # this function is called each time a record is queued to be indexed into elasticsearch.
        # This means it is called a lot and therefore needs a barrier preventing it from hammering
        # the database, hence this modulo calculation
        if indexing_stats.document_count % update_frequency == 0:
            update_stats(
                stats_id,
                {
                    ImportStats.in_progress: True,
                    ImportStats.count: indexing_stats.document_count,
                },
            )

    @indexer.finish_signal.connect_via(indexer)
    def finish(_sender, indexing_stats, stats):
        # this function is called when the indexing operation completes
        finish_operation(stats_id, indexing_stats.document_count, stats)


def mark_error(stats_id, error):
    """
    Marks the ImportStats object with the given database id as having finished with an
    error. Just the error message is stored against the ImportStats object.
    "in_progress", "duration" and "end" are also updated.

    :param stats_id: the database id of the object to update
    :param error: the exception object
    """
    start = model.Session.query(ImportStats).get(stats_id).start
    end = datetime.now()
    update_stats(
        stats_id,
        {
            ImportStats.in_progress: False,
            ImportStats.duration: (end - start).total_seconds(),
            ImportStats.end: end,
            ImportStats.error: str(
                format_exception_only(type(error), error)[-1].strip()
            ),
        },
    )


def get_all_stats(resource_id):
    """
    Retrieves and returns all the ImportStats from the database associated with the
    given resource. They are ordered by ID descending which will result in the newest
    results coming back first.

    :param resource_id: the id of the resource
    :return: a Query object which can be iterated over to retrieve all the results
    """
    return list(
        model.Session.query(ImportStats)
        .filter(ImportStats.resource_id == resource_id)
        .order_by(desc(ImportStats.id))
    )


def get_last_ingest(resource_id):
    """
    Retrieve the last ingest stat object from the database, or None if there aren't any.

    :param resource_id: the resource id
    :return: an ImportStats object or None
    """
    return (
        model.Session.query(ImportStats)
        .filter(ImportStats.resource_id == resource_id)
        .filter(ImportStats.type == INGEST)
        .order_by(ImportStats.version.desc())
        .first()
    )



import datetime
import hashlib
import random
import logging

from ckan import model
from ckan.plugins import toolkit
from sqlalchemy.exc import IntegrityError

from .schema import get_latest_query_version, hash_query
from .schema import validate_query
from .slug_words import list_one, list_two, list_three
from .utils import get_available_datastore_resources, get_resources_and_versions
from ...model.slugs import DatastoreSlug, NavigationalSlug

log = logging.getLogger(__name__)


def generate_query_hash(
    query, query_version, version, resource_ids, resource_ids_and_versions
):
    """
    Given a query and the parameters required to run it, create a unique id for it (a
    hash) and returns it. The hashing algorithm used is sha1 and the output will be a 40
    character hexidecimal string.

    :param query: the query dict
    :param query_version: the query version
    :param version: the data version
    :param resource_ids: the ids of the resources under search
    :param resource_ids_and_versions: the resource ids and specific versions to search at for them
    :return: a unique id for the query, which is a hash of the query and parameters
    """
    hash_value = hashlib.sha1()
    bits = [
        hash_query(query, query_version),
        query_version,
        version,
        # sort the resource ids to ensure stability
        sorted(resource_ids) if resource_ids is not None else None,
        # sort the resource ids and versions to ensure stability
        sorted(resource_ids_and_versions.items())
        if resource_ids_and_versions is not None
        else None,
    ]
    hash_value.update('|'.join(map(str, bits)).encode('utf-8'))
    return hash_value.hexdigest()


def generate_pretty_slug(word_lists=(list_one, list_two, list_three)):
    """
    Generate a new slug using the adjective and animal lists available. The default
    word_lists value is a trio of lists: (adjectives, adjectives, animals). This
    produces >31,000,000 unique combinations which should be more than enough! This
    function does have the potential to produce duplicate adjectives (for example,
    green-green-llama) but the chances are really low and it doesn't really matter.

    :param word_lists: a sequence of word lists to choose from
    :return: the slug
    """
    return '-'.join(map(random.choice, word_lists))


def create_slug(
    context,
    query,
    query_version,
    version=None,
    resource_ids=None,
    resource_ids_and_versions=None,
    pretty_slug=True,
    attempts=5,
):
    """
    Creates a new slug in the database and returns the saved DatastoreSlug object. If a
    slug already exists under the query hash then the existing slug entry is returned.
    In addition to the slug object, also returned is information about whether the slug
    was new or not.

    Only valid queries get a slug, otherwise we raise a ValidationError.

    Only valid resource ids included in the list will be stored, any invalid ones will be excluded.
    If a list of resource ids is provided and none of the requested resource ids are valid, then a
    ValidationError is raised.

    :param context: the context dict so that we can check the validity of any resources
    :param query: the query dict
    :param query_version: the query version in use
    :param version: the version to search at
    :param resource_ids: the resources to search (a list)
    :param resource_ids_and_versions: the resources and versions to search at (a dict)
    :param pretty_slug: whether to generate a pretty slug or just use the uuid id of the slug, by
                        default this is True
    :param attempts: how many times to try creating a pretty slug, default: 5
    :return: a 2-tuple containing a boolean indicating whether the slug object returned was newly
             created and the DatastoreSlug object itself. If we couldn't create a slug object for
             some reason then (False, None) is returned.
    """
    # only store valid queries!
    validate_query(query, query_version)

    resource_ids, resource_ids_and_versions = get_resources_and_versions(
        resource_ids, resource_ids_and_versions, version
    )

    query_hash = generate_query_hash(
        query, query_version, version, resource_ids, resource_ids_and_versions
    )

    existing_slug = (
        model.Session.query(DatastoreSlug)
        .filter(DatastoreSlug.query_hash == query_hash)
        .first()
    )

    if existing_slug is not None:
        return False, existing_slug

    while attempts:
        attempts -= 1
        new_slug = DatastoreSlug(
            query_hash=query_hash,
            query=query,
            query_version=query_version,
            version=version,
            resource_ids=resource_ids,
            resource_ids_and_versions=resource_ids_and_versions,
        )

        if pretty_slug:
            new_slug.pretty_slug = generate_pretty_slug()

        try:
            new_slug.save()
            break
        except IntegrityError:
            if pretty_slug:
                # assume this failed because of the pretty slug needing to be unique and try again
                model.Session.rollback()
                continue
            else:
                # something else has happened here
                raise
    else:
        return False, None

    return True, new_slug


def create_nav_slug(
    context, query, version=None, resource_ids=None, resource_ids_and_versions=None
):
    try:
        # clear old slugs before we make new ones
        clean_nav_slugs()
    except Exception as e:
        # if it fails, log it and move on
        log.debug(f'Cleaning nav slugs failed: {e}')

    query_version = get_latest_query_version()  # it should always be the latest version
    validate_query(query, query_version)

    resource_ids, resource_ids_and_versions = get_resources_and_versions(
        resource_ids, resource_ids_and_versions, version
    )

    query_hash = generate_query_hash(
        query, query_version, version, resource_ids, resource_ids_and_versions
    )

    existing_slug = (
        model.Session.query(NavigationalSlug)
        .filter(NavigationalSlug.query_hash == query_hash)
        .first()
    )

    if existing_slug is not None:
        return False, existing_slug

    new_slug = NavigationalSlug(
        query_hash=query_hash,
        query=query,
        resource_ids_and_versions=resource_ids_and_versions,
    )
    new_slug.save()

    return True, new_slug


def resolve_slug(slug, allow_nav=True):
    """
    Resolves the given slug and returns it if it's found, otherwise None is returned.

    :param slug: the slug
    :param allow_nav: allow resolving to a navigational slug
    :return: a DatastoreSlug object or None if the slug couldn't be found
    """
    if slug.startswith(NavigationalSlug.prefix) and allow_nav:
        try:
            # clean old slugs because we don't want old ones to continue resolving
            clean_nav_slugs()
        except Exception as e:
            # if it fails, log it and move on
            log.debug(f'Cleaning nav slugs failed: {e}')
        return (
            model.Session.query(NavigationalSlug)
            .filter(NavigationalSlug.on_slug(slug))
            .first()
        )
    return (
        model.Session.query(DatastoreSlug).filter(DatastoreSlug.on_slug(slug)).first()
    )


class DuplicateSlugException(Exception):
    pass


def reserve_slug(
    reserved_pretty_slug,
    query=None,
    query_version=None,
    version=None,
    resource_ids=None,
    resource_ids_and_versions=None,
):
    """
    This function can be used to reserve a slug using a specific string. This should
    probably only be called during this extension's initialisation via the
    datastore_reserve_slugs interface function.

    If a slug already exists in the database with the same reserved pretty slug and the same
    query parameters then nothing happens.

    If a slug already exists in the database with the same reserved pretty slug but a different
    set of query parameters then a DuplicateSlugException is raised.

    If a slug already exists in the database with the same query parameters but no reserved
    pretty slug then the reserved pretty slug is added to the slug.

    :param reserved_pretty_slug: the slug string to reserve
    :param query: the query dict
    :param query_version: the query schema version
    :param version: the version of the data
    :param resource_ids: the resource ids to search
    :param resource_ids_and_versions: the resources ids and specific versions for each to search
    :return: a DatastoreSlug object that has either been found (if it already existed), created (if
             no slug existed) or updated (if a slug existed for the query parameters, but no
             reserved query string was associated with it).
    """
    # default some parameters and then assert they are all the right types. We do this because if
    # there are problems they're going to be reported back to the developer not the user
    if query is None:
        query = {}
    if query_version is None:
        query_version = get_latest_query_version()
    assert isinstance(query, dict)
    assert isinstance(query_version, str)
    if version is not None:
        assert isinstance(version, int)
    if resource_ids is not None:
        assert isinstance(resource_ids, list)
    if resource_ids_and_versions is not None:
        assert isinstance(resource_ids_and_versions, dict)

    slug = resolve_slug(reserved_pretty_slug, False)
    if slug is not None:
        # a slug with this reserved pretty slug already exists
        return slug
    else:
        # there is no slug associated with this reserved pretty slug so let's see if there's a slug
        # using the query parameters
        query_hash = generate_query_hash(
            query, query_version, version, resource_ids, resource_ids_and_versions
        )
        slug = (
            model.Session.query(DatastoreSlug)
            .filter(DatastoreSlug.query_hash == query_hash)
            .first()
        )
        if slug is None:
            # we need to make a new slug
            context = {'ignore_auth': True}
            success, slug = create_slug(
                context,
                query,
                query_version,
                version,
                resource_ids,
                resource_ids_and_versions,
                pretty_slug=False,
            )
            if not success:
                # this should never really happen
                raise Exception('Failed to create new reserved slug')
            slug.reserved_pretty_slug = reserved_pretty_slug
            slug.save()
            return slug
        else:
            # see if we can update the slug that already exists
            if slug.reserved_pretty_slug is None:
                slug.reserved_pretty_slug = reserved_pretty_slug
                slug.save()
                return slug
            else:
                raise DuplicateSlugException(
                    f'The query parameters are already associated with a '
                    f'different slug: {slug.get_slug_string()}'
                )


def clean_nav_slugs(before=None):
    """
    Delete old/expired navigational slugs.

    :param before: a datetime object; slugs created before this time will be removed
                   (defaults to 2 days ago)
    :return: the number of deleted slugs
    """
    if before is None:
        before = datetime.datetime.utcnow() - datetime.timedelta(days=2)

    old_slugs = (
        model.Session.query(NavigationalSlug)
        .filter(NavigationalSlug.created < before)
        .all()
    )
    slug_count = len(old_slugs)
    for slug in old_slugs:
        slug.delete()

    model.Session.commit()

    return slug_count



from datetime import datetime

from ckan.model import meta, DomainObject
from ckan.model.types import make_uuid
from sqlalchemy import (
    Column,
    Table,
    BigInteger,
    UnicodeText,
    UniqueConstraint,
    DateTime,
    or_,
)
from sqlalchemy.dialects.postgresql import JSONB
from ..lib.query.schema import get_latest_query_version

# this table stores query slugs
datastore_slugs_table = Table(
    'versioned_datastore_slugs',
    meta.metadata,
    Column('id', UnicodeText, primary_key=True, default=make_uuid),
    Column('query_hash', UnicodeText, nullable=False, index=True, unique=True),
    Column('query', JSONB, nullable=False),
    Column('query_version', UnicodeText, nullable=False),
    Column('version', BigInteger, nullable=True),
    Column('resource_ids', JSONB, nullable=True),
    Column('resource_ids_and_versions', JSONB, nullable=True),
    Column('pretty_slug', UnicodeText, nullable=True, index=True, unique=True),
    Column('created', DateTime, nullable=False, default=datetime.utcnow),
    Column('reserved_pretty_slug', UnicodeText, nullable=True, index=True, unique=True),
    UniqueConstraint('id', 'pretty_slug', 'reserved_pretty_slug'),
)

# this table stores transient, temporary slugs
navigational_slugs_table = Table(
    'versioned_datastore_navigational_slugs',
    meta.metadata,
    Column('id', UnicodeText, primary_key=True, default=make_uuid),
    Column('query_hash', UnicodeText, nullable=False, index=True, unique=True),
    Column('query', JSONB, nullable=False),
    Column('resource_ids_and_versions', JSONB, nullable=False, default=dict),
    Column('created', DateTime, nullable=False, default=datetime.utcnow),
)


class DatastoreSlug(DomainObject):
    """
    Object for a datastore slug row.
    """

    def get_slug_string(self):
        """
        Returns the slug string to be used for this slug. This will be the first non-
        None value from the following attributes in this order: reserved_pretty_slug,
        pretty_slug or id.

        :return: the slug string
        """
        if self.reserved_pretty_slug is not None:
            return self.reserved_pretty_slug
        elif self.pretty_slug is not None:
            return self.pretty_slug
        else:
            return self.id

    @staticmethod
    def on_slug(slug_string):
        """
        Returns an or query that can be used to find the slug in the database with the
        passed slug string.

        :param slug_string: the slug string
        :return: an or filter
        """
        return or_(
            DatastoreSlug.id == slug_string,
            DatastoreSlug.pretty_slug == slug_string,
            DatastoreSlug.reserved_pretty_slug == slug_string,
        )


class NavigationalSlug(DomainObject):
    """
    Object for a navigational slug.
    """

    prefix = 'nav-'
    version = None

    @property
    def query_version(self):
        return get_latest_query_version()

    @property
    def resource_ids(self):
        return list(self.resource_ids_and_versions.keys())

    def get_slug_string(self):
        """
        Returns the slug string to be used for this slug.

        :return: the slug string
        """
        return NavigationalSlug.prefix + self.id

    @staticmethod
    def on_slug(slug_string):
        """
        Returns an or query that can be used to find the slug in the database with the
        passed slug string.

        :param slug_string: the slug string
        :return: a filter
        """
        return NavigationalSlug.id == slug_string[len(NavigationalSlug.prefix) :]


meta.mapper(DatastoreSlug, datastore_slugs_table)
meta.mapper(NavigationalSlug, navigational_slugs_table)



from .urls import SchemaUrl
from .utils import load_schema


class Extension(object):
    def __init__(self, location, prop_names, row_type, name, core=False):
        self.location = location
        self.prop_names = prop_names or []
        self.row_type = row_type
        self.name = name
        self.core = core

    @classmethod
    def create(cls, url: SchemaUrl, extension_schema=None, core=False):
        if extension_schema is None:
            extension_schema = load_schema(url.url)
        row_type = extension_schema.attrib['rowType']
        name = extension_schema.attrib['name']
        schema_props = extension_schema.findall('.//{*}property')
        prop_names = [p.attrib['name'] for p in schema_props]
        return cls(url, prop_names, row_type, name, core)

    @classmethod
    def load(cls, serialised):
        serialised['location'] = SchemaUrl(**serialised['location'])
        return cls(**serialised)

    def serialise(self):
        return {
            'location': self.location.__dict__,
            'prop_names': self.prop_names,
            'row_type': self.row_type,
            'name': self.name,
            'core': self.core,
        }


class Prop(object):
    def __init__(self, **kwargs):
        # preferable to instantiate with .create() class method instead
        self.name = kwargs.get('name')
        self.iri = kwargs.get('iri')
        self.prop_type = kwargs.get('prop_type', str)
        self.is_identifier = kwargs.get('is_identifier', False)
        self.domain = kwargs.get('domain')
        self.vocabulary = kwargs.get('vocabulary', [])
        self._flags = kwargs.get('flags', [])
        self.extension = kwargs.get('extension')

    @classmethod
    def create(
        cls, row_source=None, xml_element_source=None, flags=None, extension=None
    ):
        flags = flags or []
        new_prop = cls()
        new_prop.update(row_source, xml_element_source, flags, extension)
        return new_prop

    def update(
        self, row_source=None, xml_element_source=None, flags=None, extension=None
    ):
        flags = flags or []
        if row_source is None and xml_element_source is None:
            raise Exception('At least one source required.')
        if row_source is not None:
            self.name = row_source.term_localName
            self.iri = row_source.term_iri
            self.domain = row_source.organized_in
        else:
            self.name = xml_element_source.attrib['name']
            self.iri = xml_element_source.attrib.get('qualName')
            self.domain = xml_element_source.attrib.get('group')
        if xml_element_source is None:
            self.is_identifier = False
            thesaurus_url = None
        else:
            self.is_identifier = (
                xml_element_source.attrib.get('substitutionGroup')
                == 'dwc:anyIdentifier'
            )
            prop_type = xml_element_source.attrib.get('type')
            if prop_type != 'xs:string':
                self.prop_type = prop_type
            thesaurus_url = xml_element_source.attrib.get('thesaurus')
        if thesaurus_url is not None:
            thesaurus = load_schema(thesaurus_url)
            terms = thesaurus.findall('.//{*}concept')
            identifier_key = next(
                k for k in terms[0].attrib.keys() if 'identifier' in k
            )
            self.vocabulary = [t.attrib[identifier_key] for t in terms]
        else:
            self.vocabulary = []
        self._flags = [
            f.lower() for f in set(self._flags + flags) if isinstance(f, str)
        ]
        self.extension = extension

    @property
    def flags(self):
        return [f.lower() for f in self._flags]

    @property
    def core(self):
        return self.extension is None

    def serialise(self):
        return {
            'name': self.name,
            'iri': self.iri,
            'is_identifier': self.is_identifier,
            'domain': self.domain,
            'vocabulary': self.vocabulary,
            'flags': self.flags,
        }

    @classmethod
    def load(cls, serialised):
        return cls(**serialised)

    def __repr__(self):
        if self.vocabulary:
            return f'{self.name} ({len(self.vocabulary)} terms)'
        else:
            return self.name

    def has_flags(self, *flags):
        return all([f.lower() in self.flags for f in flags])


class PropCollection(dict):
    def __init__(self, props):
        if not isinstance(props, dict):
            try:
                props = {p.name: p for p in props}
            except:
                raise TypeError('Must be dict or iterable.')
        super(PropCollection, self).__init__(**props)

    def flagged(self, *flags):
        return PropCollection([v for v in self.values() if v.has_flags(*flags)])

    def serialise(self):
        return {k: v.serialise() for k, v in self.items()}

    @classmethod
    def load(cls, serialised):
        if isinstance(serialised, list):
            serialised = [Prop.load(s) for s in serialised]
        elif isinstance(serialised, dict):
            serialised = {k: Prop.load(v) for k, v in serialised.items()}
        return cls(serialised)


class Domain(object):
    def __init__(self, name, label, iri):
        self.name = name
        self.label = label
        self.iri = iri

    def props(self, schema):
        if schema is None:
            return []
        return [p for p in schema.props if p.domain == self.iri]

    def identifier(self, schema):
        return next(p for p in self.props(schema) if p.is_identifier)

    @classmethod
    def create(cls, source):
        name = source.term_localName
        label = source.label
        iri = source.term_iri
        return cls(name, label, iri)

    def serialise(self):
        return {'name': self.name, 'label': self.label, 'iri': self.iri}

    @classmethod
    def load(cls, serialised):
        return cls(**serialised)

    def __repr__(self):
        return self.name



import os
from datetime import datetime as dt, timedelta

from flask import Blueprint, jsonify

from ckan.plugins import toolkit, plugin_loaded
from ..lib.downloads.loaders import get_file_server
from ..lib.downloads.servers import DirectFileServer
from ..lib.query.slugs import create_nav_slug
from ..logic.actions.meta.arg_objects import ServerArgs
from ..model.downloads import DownloadRequest

blueprint = Blueprint(
    name='datastore_status', import_name=__name__, url_prefix='/status'
)


def get_download_details(download_id):
    dl = DownloadRequest.get(download_id)

    if dl is None:
        return toolkit.render(
            'status/download.html',
            extra_vars={'download_request': None},
        )

    res_show = toolkit.get_action('resource_show')
    resources = {}
    if dl.core_record:
        for k in dl.core_record.resource_ids_and_versions:
            try:
                resources[k] = res_show({}, {'id': k})
            except:
                continue

    status_friendly = {
        DownloadRequest.state_initial: toolkit._('Waiting to start'),
        DownloadRequest.state_core_gen: toolkit._('Searching resources'),
        DownloadRequest.state_derivative_gen: toolkit._('Generating files'),
        DownloadRequest.state_retrieving: toolkit._('Retrieving files'),
        DownloadRequest.state_complete: toolkit._('Complete'),
        DownloadRequest.state_failed: toolkit._('Failed'),
        DownloadRequest.state_packaging: toolkit._('Packaging'),
    }

    time_now = dt.utcnow()
    end_time = dl.modified if dl.state == DownloadRequest.state_complete else time_now
    total_time_elapsed = timedelta(
        seconds=round((end_time - dl.created).total_seconds())
    )
    since_last_updated = timedelta(
        seconds=round((time_now - dl.modified).total_seconds())
    )

    if dl.derivative_record.filepath:
        file_exists = os.path.exists(dl.derivative_record.filepath)
    else:
        file_exists = False

    urls = {}
    if file_exists and dl.state == DownloadRequest.state_complete:
        # include a vanilla direct link
        urls['direct'] = DirectFileServer().serve(dl)
        if dl.server_args is not None:
            server_args = ServerArgs(**dl.server_args)
            if server_args.custom_filename:
                urls['custom'] = DirectFileServer(
                    filename=server_args.custom_filename
                ).serve(dl)

            server = get_file_server(
                server_args.type,
                filename=server_args.custom_filename,
                **server_args.type_args,
            )
            url = server.serve(dl)
            if url != urls['direct'] and url != urls.get('custom'):
                urls[server.name] = url

    query_doi = None
    doi_url = None
    if plugin_loaded('query_dois'):
        from ckanext.query_dois.lib.doi import find_existing_doi
        from ckanext.query_dois.helpers import get_landing_page_url

        # query-dois only saves resources that return records
        non_zero_resources = {
            k: v
            for k, v in dl.core_record.resource_ids_and_versions.items()
            if dl.core_record.resource_totals.get(k, 0) > 0
        }

        query_doi = find_existing_doi(
            non_zero_resources,
            dl.core_record.query_hash,
            dl.core_record.query_version,
        )
        if query_doi:
            doi_url = get_landing_page_url(query_doi)

    try:
        nav_slug = create_nav_slug(
            {},
            dl.core_record.query,
            resource_ids_and_versions=dl.core_record.resource_ids_and_versions,
        )[1]
        search_url = toolkit.url_for(
            'search.view', slug=nav_slug.get_slug_string(), qualified=True
        )
    except toolkit.ValidationError:
        # if the resource is a non-datastore resource, this will fail, so just don't
        # return a search url
        search_url = None

    return {
        'download_request': dl,
        'file_exists': file_exists,
        'resources': resources,
        'status': dl.state,
        'status_friendly': status_friendly[dl.state],
        'total_time': total_time_elapsed,
        'since_last_update': since_last_updated,
        'urls': urls,
        'doi': query_doi,
        'doi_url': doi_url,
        'search_url': search_url,
    }


@blueprint.route('/download/<download_id>')
def download_status(download_id):
    details = get_download_details(download_id)
    return toolkit.render(
        'status/download.html',
        extra_vars=details,
    )


@blueprint.route('/download/<download_id>/json')
def download_status_json(download_id):
    details = get_download_details(download_id)

    download_request = details['download_request']
    details['created'] = download_request.created
    details['modified'] = download_request.modified
    details['message'] = download_request.message
    del details['download_request']  # not serialisable

    doi = details.get('doi')
    if doi:
        details['query'] = doi.query
        details['query_version'] = doi.query_version
        details['doi'] = doi.doi

    details['download_id'] = download_id
    details['total_time'] = details['total_time'].seconds
    details['since_last_update'] = details['since_last_update'].seconds
    return jsonify(details)



from .email import EmailNotifier
from .webhook import WebhookNotifier
from .null import NullNotifier
from ckan.plugins import toolkit

notifiers = [EmailNotifier, WebhookNotifier, NullNotifier]


def validate_notifier_args(notifier_type, notifier_type_args):
    try:
        notifier = next(n for n in notifiers if n.name == notifier_type)
    except StopIteration:
        raise toolkit.Invalid('Invalid notifier type.')

    return notifier.validate_args(notifier_type_args)



import click

from ckan.plugins import toolkit
from .model.details import datastore_resource_details_table
from .model.downloads import (
    datastore_downloads_requests_table,
    datastore_downloads_derivative_files_table,
    datastore_downloads_core_files_table,
)
from .model.slugs import datastore_slugs_table, navigational_slugs_table
from .model.stats import import_stats_table


def get_commands():
    return [versioned_datastore]


@click.group()
def versioned_datastore():
    """
    Perform various tasks on the versioned datastore.
    """
    pass


@versioned_datastore.command(name='initdb')
def init_db():
    """
    Ensure the tables needed by this plugin exist.
    """
    tables = [
        import_stats_table,
        datastore_resource_details_table,
        datastore_slugs_table,
        navigational_slugs_table,
        datastore_downloads_core_files_table,
        datastore_downloads_derivative_files_table,
        datastore_downloads_requests_table,
    ]
    # create the tables if they don't exist
    for table in tables:
        if not table.exists():
            table.create()
    click.secho('Finished creating tables', fg='green')


@versioned_datastore.command()
@click.option('-r', '--resource_id', 'resource_ids', multiple=True)
def reindex(resource_ids):
    """
    Reindex either a specific resource or all resources.
    """
    ids = set()
    context = {'ignore_auth': True}

    if resource_ids:
        # the user has specified some resources to reindex
        ids.update(resource_ids)
    else:
        # the user hasn't specified the resources to reindex so we should get a list of all
        # resources in the system
        data_dict = {'query': 'name:', 'offset': 0, 'limit': 100}
        while True:
            result = toolkit.get_action('resource_search')(context, data_dict)
            if len(result['results']) > 0:
                ids.update(resource['id'] for resource in result['results'])
                data_dict['offset'] += data_dict['limit']
            else:
                break

    if not ids:
        click.secho('No resources found to reindex', fg='green')
        return

    click.secho(f'Found {len(ids)} resources to reindex', fg='yellow')

    for resource_id in sorted(ids):
        try:
            result = toolkit.get_action('datastore_reindex')(
                context, {'resource_id': resource_id}
            )
            click.secho(
                f'Queued reindex of {resource_id} as job {result["job_id"]}', fg='cyan'
            )
        except toolkit.ValidationError as e:
            click.secho(
                f'Failed to reindex {resource_id} due to validation error: {e}',
                fg='red',
            )

    click.secho('Reindexing complete', fg='green')



import hashlib

from ckan.plugins import toolkit
from ..basic_query.utils import convert_to_multisearch
from ..query.schema import (
    get_latest_query_version,
    hash_query,
    translate_query,
    validate_query,
)
from ..query.utils import get_resources_and_versions
from ...logic.actions.meta.arg_objects import QueryArgs


class Query(object):
    def __init__(self, query, query_version, resource_ids_and_versions):
        self.query = query
        self.query_version = query_version
        self.resource_ids_and_versions = resource_ids_and_versions
        self._search = None
        self.validate()

    @property
    def search(self):
        if self._search is None:
            self._search = self.translate()
        return self._search

    @classmethod
    def from_query_args(cls, query_args: QueryArgs, allow_non_datastore=False):
        query = query_args.query
        query_version = query_args.query_version
        resource_ids = query_args.resource_ids
        resource_ids_and_versions = query_args.resource_ids_and_versions
        version = query_args.version

        if query_args.slug_or_doi:
            try:
                saved_query = toolkit.get_action('datastore_resolve_slug')(
                    {}, {'slug': query_args.slug_or_doi}
                )
                query = saved_query.get('query')
                query_version = saved_query.get('query_version')
                resource_ids = saved_query.get('resource_ids')
                resource_ids_and_versions = saved_query.get('resource_ids_and_versions')
            except toolkit.ValidationError:
                # if the slug doesn't resolve, continue as normal
                pass

        resource_ids, rounded_resource_ids_and_versions = get_resources_and_versions(
            resource_ids,
            resource_ids_and_versions,
            allow_non_datastore=allow_non_datastore,
        )

        # setup the query
        if query is None:
            query = {}
        if query_version and query_version.lower().startswith('v0'):
            # this is an old/basic query so we need to convert it first
            query = convert_to_multisearch(query)
            query_version = None
        if query_version is None:
            query_version = get_latest_query_version()

        return cls(query, query_version, rounded_resource_ids_and_versions)

    @property
    def hash(self):
        return hash_query(self.query, self.query_version)

    @property
    def resource_hash(self):
        resources = sorted(self.resource_ids_and_versions.items())
        return hashlib.sha1('|'.join(map(str, resources)).encode('utf-8')).hexdigest()

    @property
    def record_hash(self):
        to_hash = [self.hash, self.resource_hash]
        download_hash = hashlib.sha1('|'.join(to_hash).encode('utf-8'))
        return download_hash.hexdigest()

    def validate(self):
        return validate_query(self.query, self.query_version)

    def translate(self):
        return translate_query(self.query, self.query_version)



import json
import os

import pandas as pd

from .schema_parts import Domain, Extension, Prop, PropCollection
from .urls import TDWGUrls
from .utils import load_schema


class Schema(object):
    def __init__(
        self,
        domains,
        props,
        core_extension=None,
        extensions=None,
        extension_props=None,
        row_type='http://rs.tdwg.org/dwc/terms/Occurrence',
    ):
        self.domains = domains
        self.props = props
        self.core_extension = core_extension
        self.extensions = extensions or []
        self.extension_props = extension_props or PropCollection([])
        self.row_type = core_extension.row_type if core_extension else row_type
        self.row_type_name = (
            core_extension.name if core_extension else row_type.split('/')[-1]
        )

    @classmethod
    def regenerate(cls, core_extension_url=None, extension_urls=None):
        extension_urls = extension_urls or []

        df = pd.read_csv(TDWGUrls.terms_csv)
        # ignore deprecated terms
        df = df[df.status == 'recommended']
        # drop some columns we don't need
        df = df.drop(['issued', 'status', 'replaces'], axis=1)

        domains_df = df[df.rdf_type == 'http://www.w3.org/2000/01/rdf-schema#Class']
        domains_df = domains_df.drop(['organized_in', 'rdf_type'], axis=1)
        domains = [Domain.create(d) for _, d in domains_df.iterrows()]

        props_df = df[
            df.rdf_type == 'http://www.w3.org/1999/02/22-rdf-syntax-ns#Property'
        ]
        props_df = props_df.drop(['rdf_type'], axis=1)

        # supplementary property information is in the core schema
        core_xsd = load_schema(TDWGUrls.core, base_url=TDWGUrls.base_url)
        core_props = {}
        for _, p in props_df.iterrows():
            xml_element = core_xsd.find(f'.//{{*}}element[@name="{p.term_localName}"]')
            # (hopefully) temporary fix for https://github.com/tdwg/dwc/issues/403
            if xml_element is None and p.term_localName == 'degreeOfEstablishment':
                xml_element = core_xsd.find(
                    './/{*}element[@name="degreeOfEstablishmentMeans"]'
                )
            elif xml_element is None and p.term_localName == 'waterBody':
                xml_element = core_xsd.find('.//{*}element[@name="waterbody"]')
            prop = Prop.create(p, xml_element, flags=[p['flags']])
            core_props[prop.name] = prop

        if core_extension_url is not None:
            temp_props = {}
            extension_schema = load_schema(
                core_extension_url.url, core_extension_url.base
            )
            ce_props = extension_schema.findall('.//{*}property')
            for p in ce_props:
                domain_name = p.attrib['group']
                prop_name = p.attrib['name']
                prop = core_props.get(prop_name)
                if prop is not None:
                    prop.update(xml_element_source=p)
                else:
                    prop = Prop.create(xml_element_source=p, flags=['core_extension'])
                temp_props[prop.name] = prop
            core_props = temp_props
            ce = Extension.create(core_extension_url, extension_schema, core=True)
        else:
            ce = None

        props = PropCollection(core_props.values())

        extensions = []
        extension_props = {}
        for e in extension_urls:
            extension_schema = load_schema(e.url, e.base)
            ext = Extension.create(e, extension_schema)
            extensions.append(ext)
            schema_props = extension_schema.findall('.//{*}property')
            extension_props[ext.name] = PropCollection(
                [
                    Prop.create(xml_element_source=p, extension=ext.name)
                    for p in schema_props
                ]
            )

        return cls(domains, props, ce, extensions, extension_props)

    def serialise(self):
        serialised_dict = {
            'domains': [d.serialise() for d in self.domains],
            'props': self.props.serialise(),
            'row_type': self.row_type,
        }
        if self.core_extension is not None:
            serialised_dict['core_extension'] = self.core_extension.serialise()
        if len(self.extensions) > 0:
            serialised_dict['extensions'] = [e.serialise() for e in self.extensions]
            serialised_dict['extension_props'] = {
                k: v.serialise() for k, v in self.extension_props.items()
            }
        return serialised_dict

    def save(self, cache_path):
        with open(cache_path, 'w') as f:
            json.dump(self.serialise(), f, indent=2)

    @classmethod
    def load(cls, cache_path=None, **kwargs):
        if cache_path is None:
            return cls.regenerate(**kwargs)
        if os.path.exists(cache_path):
            with open(cache_path, 'r') as f:
                content = json.load(f)
            init_dict = {
                'domains': [Domain.load(d) for d in content['domains']],
                'props': PropCollection.load(content['props']),
                'extensions': [
                    Extension.load(e) for e in content.get('extensions', [])
                ],
                'extension_props': {
                    k: PropCollection.load(v)
                    for k, v in content.get('extension_props', {}).items()
                },
                'row_type': content.get(
                    'row_type', 'http://rs.tdwg.org/dwc/terms/Occurrence'
                ),
            }
            ce = content.get('core_extension')
            if ce is not None:
                init_dict['core_extension'] = Extension.load(ce)
                init_dict['row_type'] = init_dict['core_extension'].row_type
            return cls(**init_dict)
        else:
            new_archiver = cls.regenerate(**kwargs)
            new_archiver.save(cache_path)
            return new_archiver



import csv
import json
import os
import shutil
from collections import OrderedDict
from datetime import datetime as dt
from uuid import uuid4

from lxml import etree

from ckan.plugins import toolkit, plugin_loaded, PluginImplementations
from . import urls, utils
from .schema import Schema
from ..base import BaseDerivativeGenerator
from .....interfaces import IVersionedDatastoreDownloads


class DwcDerivativeGenerator(BaseDerivativeGenerator):
    name = 'dwc'
    extension = 'zip'

    def __init__(
        self,
        output_dir,
        fields,
        query,
        resource_id=None,
        core_extension_name=None,
        extension_names=None,
        extension_map=None,
        id_field='_id',
        eml_title=None,
        eml_abstract=None,
        **format_args,
    ):
        super(DwcDerivativeGenerator, self).__init__(
            output_dir, fields, query, resource_id, **format_args
        )
        self._id_field = id_field
        self._eml_title = eml_title
        self._eml_abstract = eml_abstract
        self._build_dir_name = uuid4().hex
        self._build_dir = os.path.join(self.output_dir, self._build_dir_name)
        os.mkdir(self._build_dir)

        # LOAD THE SCHEMA
        schema_args = {}
        # there can be maximum one core extension
        core_extension_name = core_extension_name or toolkit.config.get(
            'ckanext.versioned_datastore.dwc_core_extension_name'
        )
        if (
            core_extension_name is not None
            and core_extension_name.lower() in urls.core_extensions
        ):
            core_ext_url = urls.core_extensions.get(core_extension_name.lower())
            schema_args['core_extension_url'] = core_ext_url
        # there can be multiple non-core extensions, separated by commas in ckan.ini or supplied as
        # a list in the request args
        if extension_names is None:
            extension_names = [
                e.strip().lower()
                for e in toolkit.config.get(
                    'ckanext.versioned_datastore.dwc_extension_names', ''
                ).split(',')
            ]
        # the fields used by the extension can also be overridden by request args
        extension_map = extension_map or {}
        ext_urls = []
        for e in extension_names:
            ext = urls.extensions.get(e)
            if not ext:
                continue
            fields = extension_map.get(e)
            if fields is not None and isinstance(fields, list):
                ext.fields = fields
            if ext.fields:  # no point in adding it if no fields are defined
                ext_urls.append(ext)
        if len(ext_urls) > 0:
            schema_args['extension_urls'] = ext_urls
        self.schema = Schema.load(
            toolkit.config.get('ckanext.versioned_datastore.dwc_schema_cache'),
            **schema_args,
        )

        # set up file paths & divide fields between core/extensions
        self._core_file_name = f'{self.schema.row_type_name.lower()}.csv'
        self.file_paths = {'core': os.path.join(self._build_dir, self._core_file_name)}
        root_field_names = set([f.split('.')[0] for f in self.all_fields])
        all_ext_field_names = [
            item for ext in self.schema.extensions for item in ext.location.fields
        ]
        dataset_fields = [
            f
            for f in root_field_names
            if f not in all_ext_field_names and f in self.schema.props
        ]
        self.fields = {
            'core': ['_id'] + sorted(list(set(utils.standard_fields + dataset_fields)))
        }

        for e in self.schema.extensions:
            self.file_paths[e.name] = os.path.join(
                self._build_dir, f'{e.name.lower()}.csv'
            )
            subfields = [
                f.split('.')[-1]
                for f in self.all_fields
                if f.split('.')[0] in e.location.fields
            ]
            potential_fields = self.schema.extension_props[e.name]
            ext_field_names = ['_id'] + list(
                set([f for f in subfields if f in potential_fields])
            )
            self.fields[e.name] = ext_field_names
        # this will contain csv writers for each component csv
        self.writers = {}
        # add meta and eml paths
        self.file_paths['_meta'] = os.path.join(self._build_dir, 'meta.xml')
        self.file_paths['_eml'] = os.path.join(self._build_dir, 'eml.xml')

        # counts number of rows written
        self.rows_written = 0

    def __exit__(self, *args, **kwargs):
        super(DwcDerivativeGenerator, self).__exit__(*args, **kwargs)
        shutil.make_archive(
            os.path.join(self.output_dir, os.path.splitext(self.output_name)[0]),
            'zip',
            self._build_dir,
        )

    def setup(self):
        if not self._opened:
            raise Exception('Files should be open.')
        self.writers = {
            k: csv.DictWriter(self.files[k], self.fields[k], dialect='unix')
            for k in self.fields
        }
        # headers are written in .validate()
        super(DwcDerivativeGenerator, self).setup()

    def validate(self, record):
        if 'type' in record and record['type'] not in utils.valid_types:
            self.writers['core'].fieldnames = [
                f for f in self.fields['core'] if f != 'type'
            ]
        for k, writer in self.writers.items():
            writer.writeheader()
        super(DwcDerivativeGenerator, self).validate(record)

    def finalise(self):
        self.files['_meta'].write(self.make_meta())
        self.files['_eml'].write(self.make_eml())
        self.writers = {}

    def cleanup(self):
        try:
            shutil.rmtree(self._build_dir)
        except FileNotFoundError:
            pass

    def _write(self, record):
        core_row, ext_rows = self._extract_record(record)
        self.writers['core'].writerow(core_row)
        for e, rows in ext_rows.items():
            for row in rows:
                self.writers[e].writerow(row)
        self.rows_written += 1

    def _extract_record(self, record):
        """
        Transform the data in a single row into the required format. Separates extension
        fields from core fields and filters out any fields that don't match the schema.

        :param record: the row of data
        :return: core row (dict), extension rows keyed on extension name (dict of lists of dicts)
        """
        core = {}
        ext = {}
        dynamic_properties = {}

        if self._id_field not in record:
            raise Exception(f'Record does not have ID field {self._id_field}')
        record_id = record.get(self._id_field)
        core['_id'] = record_id

        extension_map = {}
        for e in self.schema.extensions:
            for f in e.location.fields:
                extension_map[f] = e.name

        for k, v in record.items():
            if k in extension_map:
                ext_props = self.writers[extension_map[k]].fieldnames

                def _extract_ext(subdict):
                    props = {'_id': record_id}
                    for ek, ev in subdict.items():
                        if ek in ext_props and ek != '_id':
                            props[ek] = ev
                    return props

                if isinstance(v, list):
                    ext_extracted = [_extract_ext(x) for x in v]
                elif isinstance(v, dict):
                    ext_extracted = [_extract_ext(v)]
                else:
                    ext_extracted = [_extract_ext({k: v})]
                ext[extension_map[k]] = ext_extracted
            else:
                if k in self.writers['core'].fieldnames:
                    core[k] = v
                else:
                    dynamic_properties[k] = v

        core['dynamicProperties'] = json.dumps(dynamic_properties)
        return core, ext

    def make_meta(self):
        """
        Create the xml text content of the metafile.

        :return: xml string
        """
        nsmap = utils.NSMap(xsi=urls.XMLUrls.xsi, xs=urls.XMLUrls.xs)
        root = etree.Element(
            'archive',
            nsmap=nsmap,
            attrib={
                'xmlns': urls.TDWGUrls.xmlns,
                'metadata': 'eml.xml',
                nsmap.ns('xsi', 'schemaLocation'): ' '.join(
                    [urls.TDWGUrls.xmlns, urls.TDWGUrls.metadata]
                ),
            },
        )
        attributes = {
            'encoding': 'UTF-8',
            'linesTerminatedBy': '\n',
            'fieldsTerminatedBy': ',',
            'fieldsEnclosedBy': '"',
            'ignoreHeaderLines': '1',
        }
        core = etree.SubElement(
            root, 'core', rowType=self.schema.row_type, **attributes
        )
        core_files = etree.SubElement(core, 'files')
        core_files_location = etree.SubElement(core_files, 'location')
        core_files_location.text = self._core_file_name
        etree.SubElement(core, 'id', index='0')
        for i, c in enumerate(self.writers['core'].fieldnames):
            if c == '_id':
                continue
            prop = self.schema.props.get(c)
            if prop is None:
                continue
            etree.SubElement(core, 'field', index=str(i), term=prop.iri)
        for e in self.schema.extensions:
            ext_root = etree.SubElement(
                root, 'extension', rowType=e.row_type, **attributes
            )
            ext_files = etree.SubElement(ext_root, 'files')
            ext_files_location = etree.SubElement(ext_files, 'location')
            ext_files_location.text = f'{e.name.lower()}.csv'
            etree.SubElement(ext_root, 'coreid', index='0')
            ext_props = self.schema.extension_props[e.name]
            for i, c in enumerate(self.writers[e.name].fieldnames):
                if c == '_id':
                    continue
                etree.SubElement(
                    ext_root, 'field', index=str(i), term=ext_props.get(c).iri
                )
        return etree.tostring(root, pretty_print=True).decode()

    def make_eml(self):
        """
        Create the xml text content of the resource metadata file.

        Tries to use some sensible
        defaults and get information from other relevant plugins where available, but there's still
        the potential for errors or silly data.
        :return: xml string
        """
        # load some useful actions so we don't have to fetch them repeatedly
        resource_show = toolkit.get_action('resource_show')
        package_show = toolkit.get_action('package_show')

        # get the resources and packages associated with the query
        resources = [
            resource_show({}, {'id': r}) for r in self._query.resource_ids_and_versions
        ]
        packages = [package_show({}, {'id': r['package_id']}) for r in resources]

        # useful bools
        single_resource = len(resources) == 1
        single_package = len(packages) == 1
        empty_query = self._query.query == {}

        # define some site variables
        site_org = {
            'organizationName': utils.get_setting(
                'ckanext.versioned_datastore.dwc_org_name',
                'ckanext.doi.publisher',
                'ckan.site_title',
            ),
            'electronicMailAddress': utils.get_setting(
                'ckanext.versioned_datastore.dwc_org_email', 'smtp.mail_from'
            ),
            'onlineUrl': utils.get_setting('ckan.site_url'),
        }
        site_name = utils.get_setting('ckanext.doi.site_title', 'ckan.site_title')
        site_url = utils.get_setting('ckan.site_url', default='')

        # get the license
        licenses = list(
            set(
                [
                    p.get('license_title')
                    for p in packages
                    if p.get('license_title') is not None
                ]
            )
        )
        query_license = (
            licenses[0]
            if len(licenses) == 1
            else utils.get_setting(
                'ckanext.versioned_datastore.dwc_default_license', default='null'
            )
        )

        # generate a string representing the records' location, e.g. a single resource
        # name, a package name, or the whole site
        if single_resource:
            container_name = resources[0]['name']
        elif single_package:
            container_name = f'{len(resources)} resources in {packages[0]["title"]}'
        else:
            container_name = site_name

        # generate title and abstract
        query_title = self._eml_title or f'Query on {container_name}'
        query_abstract = (
            self._eml_abstract
            or f'Query ID {self._query.hash} on {container_name} ({self.rows_written} records).'
        )

        # set up the metadata dict in order (otherwise GBIF complains) with some defaults
        dataset_metadata = OrderedDict(
            {
                'alternateIdentifier': [self._query.hash],
                'title': query_title,
                'creator': [site_org],
                'metadataProvider': [site_org],
                'pubDate': dt.now().strftime('%Y-%m-%d'),
                'language': utils.get_setting('ckan.locale_default'),
                'abstract': {'para': query_abstract},
                'keywordSet': [],
                'intellectualRights': {'para': query_license},
                'distribution': (
                    {'online': {'url': (site_url, {'function': 'information'})}},
                    {'scope': 'document'},
                ),
                'coverage': {
                    'geographicCoverage': {
                        'geographicDescription': 'Unbound',
                        'boundingCoordinates': {
                            'westBoundingCoordinates': -180,
                            'eastBoundingCoordinates': 180,
                            'northBoundingCoordinates': -90,
                            'southBoundingCoordinates': 90,
                        },
                    }
                },
                'contact': [site_org],
            }
        )
        additional_metadata = {
            'metadata': {
                'gbif': {
                    'dateStamp': dt.now().strftime('%Y-%m-%dT%H:%M:%SZ'),
                    'citation': None,
                    'resourceLogoUrl': site_url
                    + toolkit.config.get('ckan.site_logo', ''),
                }
            }
        }

        if single_package and empty_query:
            # all the resources are in a single package (even if not all resources in
            # that package are included), so link to the package
            dataset_metadata['distribution'][0]['online']['url'] = (
                toolkit.url_for('package.read', id=packages[0]['id'], qualified=True),
                {'function': 'information'},
            )

            # also link to the package via the doi
            if 'doi' in packages[0]:
                dataset_metadata['alternateIdentifier'].append(
                    'doi:' + packages[0]['doi']
                )
                pkg_year = toolkit.h.package_get_year(packages[0])
                pkg_title = packages[0]['title']
                pkg_doi = packages[0]['doi']
                pkg_author = packages[0]['author']
                additional_metadata['metadata']['gbif']['citation'] = (
                    f'{pkg_author} ({pkg_year}). Dataset: {pkg_title}. {site_name}',
                    {'identifier': f'https://doi.org/{pkg_doi}'},
                )

            if single_resource and empty_query:
                # if it's just downloading one resource with no filters, we can be even
                # more specific with some of the metadata
                resource = resources[0]
                package = packages[0]
                dataset_metadata['title'] = resource['name']
                dataset_metadata['alternateIdentifier'] += [resource['id']]
                dataset_metadata['pubDate'] = package.get(
                    'metadata_modified', resource.get('created')
                )
                dataset_metadata['abstract']['para'] = resource.get('description')
                dataset_metadata['distribution'][0]['online']['url'] = (
                    toolkit.url_for(
                        'resource.read',
                        id=package['id'],
                        resource_id=resource['id'],
                        qualified=True,
                    ),
                    {'function': 'information'},
                )
        elif plugin_loaded('query_dois'):
            from ckanext.query_dois.lib.doi import find_existing_doi

            query_doi = find_existing_doi(
                self._query.resource_ids_and_versions,
                self._query.hash,
                self._query.query_version,
            )
            if query_doi:
                dataset_metadata['alternateIdentifier'] += ['doi:' + query_doi.doi]
                dataset_metadata['distribution'][0]['online']['url'] = (
                    toolkit.url_for(
                        'query_doi.landing_page',
                        data_centre=utils.get_setting('ckanext.query_dois.prefix'),
                        identifier=query_doi.doi.split('/', 1)[-1],
                        qualified=True,
                    ),
                    {'function': 'information'},
                )
                additional_metadata['metadata']['gbif']['citation'] = (
                    toolkit.h.create_multisearch_citation_text(query_doi, html=False),
                    {'identifier': f'https://doi.org/{query_doi.doi}'},
                )

        if (
            self.schema.core_extension
            and self.schema.core_extension.location.base == urls.GBIFUrls.base_url
        ):
            dataset_metadata['keywordSet'].append(
                {
                    'keyword': self.schema.row_type_name,
                    'keywordThesaurus': f'GBIF Dataset Type Vocabulary: {urls.GBIFUrls.thesaurus}',
                }
            )

        if plugin_loaded('attribution'):
            package_contributions_show = toolkit.get_action(
                'package_contributions_show'
            )
            authors = []
            for r in resources:
                contributions = package_contributions_show({}, {'id': r['package_id']})
                for c in contributions['contributions']:
                    if c['agent']['id'] in authors:
                        continue
                    agent = {}
                    if c['agent']['external_id']:
                        agent['userId'] = (
                            c['agent']['external_id'],
                            {'directory': c['agent']['external_id_url']},
                        )
                    if c['agent']['agent_type'] == 'person':
                        agent['individualName'] = {
                            'givenName': c['agent']['given_names'],
                            'surName': c['agent']['family_name'],
                        }
                    else:
                        agent['organizationName'] = c['agent']['name']
                    dataset_metadata['creator'].append(agent)
                    authors.append(c['agent']['id'])

        for plugin in PluginImplementations(IVersionedDatastoreDownloads):
            dataset_metadata = plugin.download_modify_eml(dataset_metadata, self._query)

        nsmap = utils.NSMap(
            eml=urls.XMLUrls.eml,
            dc=urls.XMLUrls.dc,
            xsi=urls.XMLUrls.xsi,
            xml=urls.XMLUrls.xml,
        )

        root = etree.Element(
            nsmap.ns('eml', 'eml'),
            nsmap=nsmap,
            attrib={
                'scope': 'system',
                'system': 'http://gbif.org',
                nsmap.ns('xsi', 'schemaLocation'): ' '.join(
                    [urls.XMLUrls.eml, urls.GBIFUrls.eml]
                ),
                'packageId': self.output_name,
                nsmap.ns('xml', 'lang'): 'en',
            },
        )

        for i in utils.json_to_xml('dataset', dataset_metadata):
            root.append(i)

        if additional_metadata['metadata']['gbif'].get('citation') is not None:
            for i in utils.json_to_xml('additionalMetadata', additional_metadata):
                root.append(i)

        return etree.tostring(root, pretty_print=True).decode()



from collections import defaultdict, namedtuple

from mock import patch, MagicMock

from tests.helpers.utils import sync_enqueue_job


def rounded_versions():
    """
    Patches get_rounded_versions to return the target version regardless of input.

    :return:
    """

    def rounded_versions_mock(indices, target_version):
        return defaultdict(lambda: target_version)

    return patch(
        'ckanext.versioned_datastore.lib.common.SEARCH_HELPER.get_rounded_versions',
        side_effect=rounded_versions_mock,
    )


def elasticsearch_scan():
    """
    Patches elasticsearch_dsl.Search.scan to return a test resource.
    """
    MockHit = namedtuple('MockHit', ['name', 'data'])
    return patch(
        'ckanext.versioned_datastore.lib.query.utils.Search.scan',
        return_value=[
            MockHit(
                name='test-resource-id',
                data=MagicMock(
                    **{'to_dict.return_value': {'scientificName': 'Boops boops'}}
                ),
            )
        ],
    )


def enqueue_job():
    return patch('ckan.plugins.toolkit.enqueue_job', side_effect=sync_enqueue_job)


def elasticsearch_client():
    return patch(
        'ckanext.versioned_datastore.lib.downloads.download.get_elasticsearch_client',
        side_effect=MagicMock(),
    )


def get_available_resources(resource_ids=None):
    resource_ids = resource_ids or ['test-resource-id']
    return patch(
        'ckanext.versioned_datastore.lib.query.utils.get_available_datastore_resources',
        return_value=resource_ids,
    )


def query_schemas():
    test_schemas = {'v1.0.0': MagicMock(validate=MagicMock(return_value=True))}
    return patch('ckanext.versioned_datastore.lib.query.schema.schemas', test_schemas)


def url_for():
    return patch('ckan.plugins.toolkit.url_for', return_value='/banana')



import gzip
import hashlib
import sqlite3
import tempfile
from contextlib import contextmanager, closing

import requests


def compute_hash(fp):
    """
    Given a file pointer like object, computes the sha1 hash of its data. The file
    pointer is reset after use.

    :param fp: a file pointer like object
    :return: the sha1 hexdigest
    """
    hasher = hashlib.sha1()
    with ensure_reset(fp):
        while True:
            chunk = fp.read(16384)
            if not chunk:
                break
            hasher.update(chunk)
    return hasher.hexdigest()


@contextmanager
def download_to_temp_file(url, headers=None, compress=True, chunk_size=1024):
    """
    Streams the data from the given URL and saves it in a temporary file. The (named)
    temporary file is then yielded to the caller for use. Once the context collapses the
    temporary file is removed.

    If the compress parameter is passed as True (the default) the data will be downloaded and
    written to a file using GZIP and a GzipFile pointer will be returned.

    :param url: the url to stream the data from
    :param headers: a dict of headers to pass with the request
    :param compress: whether to compress the downloaded data when storing it, if so a GzipFile
                     pointer will be returned (default: True)
    :param chunk_size: the number of bytes to read at a time from the url stream
    """
    headers = headers if headers else {}
    # open up the url for streaming
    with closing(requests.get(url, stream=True, headers=headers)) as r:
        # check that we got a response we can use!
        r.raise_for_status()

        def download(fp):
            # iterate over the data from the url stream in chunks
            for chunk in r.iter_content(chunk_size=chunk_size):
                # only write chunks with data in them
                if chunk:
                    # write the chunk to the file
                    fp.write(chunk)

        # create a temporary file to store the data in
        with tempfile.NamedTemporaryFile(delete=True) as temp:
            if compress:
                with gzip.open(temp.name, mode='wb') as f:
                    download(f)
                with gzip.open(temp.name, mode='rb') as g:
                    yield g
            else:
                download(temp)
                temp.seek(0)
                yield temp


@contextmanager
def ensure_reset(file_pointer):
    """
    Context manager which resets (seeks to 0) the passed file pointer after use.

    :param file_pointer: the file pointer to reset
    """
    try:
        yield
    finally:
        file_pointer.seek(0)


class InclusionTracker(object):
    """
    Class that tracks the ids of records that have been included in an ingestion event
    but we're actually modified. The ids are stored in a temporary sqlite database in
    case there are too many to store in memory. The objects created from this class
    definition should be used as context managers like so:

    with InclusionTracker(ingester) as tracker:     ...
    """

    def __init__(self, ingester):
        '''
        :param ingester: the ingester object - we'll use the update signal to track the ids
        '''
        self.ingester = ingester
        self.temporary_file = None
        self.tracker_db = None

    def __enter__(self):
        self.temporary_file = tempfile.NamedTemporaryFile(delete=True)
        self.tracker_db = sqlite3.connect(self.temporary_file.name)
        self.tracker_db.execute('create table ids (id integer primary key)')

        @self.ingester.update_signal.connect_via(self.ingester)
        def on_update(_sender, record, doc):
            # this implies that the record update wasn't written to mongo because the record
            # hasn't changed in this new version
            if not doc:
                self.tracker_db.execute('insert into ids(id) values (?)', (record.id,))

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.tracker_db.commit()
        self.tracker_db.close()
        self.temporary_file.close()

    def was_included(self, record_id):
        """
        Checks whether the given record id was included in the ingestion but not updated
        as the data was the same.

        :param record_id: the record id
        :return: True if the record id was included but not updated and False if not
        """
        with self.tracker_db:
            cursor = self.tracker_db.cursor()
            cursor.execute('select 1 from ids where id = ?', (record_id,))
            return cursor.fetchone() is not None



from abc import ABCMeta, abstractmethod


class BaseTransform(metaclass=ABCMeta):
    """
    A factory class for transforming data from a core file.
    """

    name = 'base'

    def __init__(self, **kwargs):
        self.transformer_args = kwargs

    @abstractmethod
    def __call__(self, data):
        return data



from collections import defaultdict
from datetime import datetime
from typing import Dict

import jsonschema
from ckantools.decorators import action
from ckantools.timer import Timer
from elasticsearch_dsl import MultiSearch, A
from splitgill.utils import to_timestamp

from ckan.plugins import toolkit, PluginImplementations, plugin_loaded
from .meta import help, schema
from ...interfaces import IVersionedDatastore
from ...lib import common
from ...lib.basic_query.utils import convert_to_multisearch
from ...lib.datastore_utils import (
    prefix_resource,
    unprefix_index,
    iter_data_fields,
    trim_index_name,
    prefix_field,
)
from ...lib.query.fields import (
    get_all_fields,
    select_fields,
    get_single_resource_fields,
    get_mappings,
)
from ...lib.query.query_log import log_query
from ...lib.query.schema import (
    get_latest_query_version,
    InvalidQuerySchemaVersionError,
    validate_query,
    translate_query,
    hash_query,
)
from ...lib.query.slugs import create_slug, resolve_slug, create_nav_slug
from ...lib.query.utils import (
    get_available_datastore_resources,
    determine_resources_to_search,
    determine_version_filter,
    calculate_after,
    find_searched_resources,
)


@action(
    schema.datastore_multisearch(), help.datastore_multisearch, toolkit.side_effect_free
)
def datastore_multisearch(
    context,
    query=None,
    query_version=None,
    version=None,
    resource_ids=None,
    resource_ids_and_versions=None,
    size=100,
    after=None,
    top_resources=False,
    timings=False,
):
    """
    Performs a search across multiple resources at the same time and returns the results
    in descending _id and index name order (the index name is included to ensure unique
    sorting otherwise the after value based pagination won't work properly).

    :param context: the context dict from the action call
    :param query: the query dict. If None (default) then an empty query is used
    :param query_version: the version of the query schema the query is using. If None (default) then
                          the latest query schema version is used
    :param version: the version to search the data at. If None (default) the current time is used
    :param resource_ids: the list of resource to search. If None (default) then all the resources
                         the user has access to are queried. If a list of resources are passed then
                         any resources not accessible to the user will be removed before querying
    :param resource_ids_and_versions: a dict of resources and versions to search each of them at.
                                      This allows precise searching of each resource at a specific
                                      parameter. If None (default) then the resource_ids parameter
                                      is used together with the version parameter. If this parameter
                                      is provided though, it takes priority over the resource_ids
                                      and version parameters.
    :param size: the number of records to return. Defaults to 100 if not provided and must be
                 between 0 and 1000.
    :param after: pagination after value that has come from a previous result. If None (default)
                  this parameter is ignored.
    :param top_resources: whether to include information about the resources with the most results
                          in them (defaults to False) in the result
    :param timings: whether to include timing information in the result dict
    :return: a dict of results including the records and total
    """
    # provide some more complex defaults for some parameters if necessary
    if query is None:
        query = {}
    if query_version is None:
        query_version = get_latest_query_version()
    size = max(0, min(size, 1000))

    timer = Timer()

    try:
        # validate and translate the query into an elasticsearch-dsl Search object
        validate_query(query, query_version)
        timer.add_event('validate')
        search = translate_query(query, query_version)
        timer.add_event('translate')
    except (jsonschema.ValidationError, InvalidQuerySchemaVersionError) as e:
        raise toolkit.ValidationError(e.message)

    # figure out which resources we're searching
    resource_ids, skipped_resource_ids = determine_resources_to_search(
        context, resource_ids, resource_ids_and_versions
    )
    timer.add_event('determine_resources')
    if not resource_ids:
        raise toolkit.ValidationError(
            "The requested resources aren't accessible to this user"
        )

    # add the version filter necessary given the parameters and the resources we're searching
    version_filter = determine_version_filter(
        version, resource_ids, resource_ids_and_versions
    )
    search = search.filter(version_filter)
    timer.add_event('version_filter')

    # add a simple default sort to ensure we get an after value for pagination. We use a combination
    # of the modified date, id of the record and the index it's in so that we get a unique sort
    search = search.sort(
        # not all indexes have a modified field so we need to provide the unmapped_type option
        {'data.modified': {'order': 'desc', 'unmapped_type': 'date'}},
        {'data._id': 'desc'},
        {'_index': 'desc'},
    )
    # add the after if there is one
    if after is not None:
        search = search.extra(search_after=after)
    # add the size parameter. We pass the requested size + 1 to allow us to determine if the results
    # we find represent the last page of results or not
    search = search.extra(size=size + 1)
    # add the resource indexes we're searching on
    search = search.index(
        [prefix_resource(resource_id) for resource_id in resource_ids]
    )

    if top_resources:
        # gather the number of hits in the top 10 most frequently represented indexes if requested
        search.aggs.bucket('indexes', 'terms', field='_index')

    # create a multisearch for this one query - this ensures there aren't any issues with the length
    # of the URL as the index list is passed as a part of the body
    multisearch = MultiSearch(using=common.ES_CLIENT).add(search)
    timer.add_event('search_params')

    # run the search and get the only result from the search results list
    result = next(iter(multisearch.execute()))
    timer.add_event('run')

    hits, next_after = calculate_after(result, size)

    response = {
        'total': result.hits.total,
        'after': next_after,
        'records': [
            {
                'data': hit.data.to_dict(),
                # should we provide the name too? If so cache a map of id -> name, then update it if we
                # don't find the id in the map
                'resource': trim_index_name(hit.meta.index),
            }
            for hit in hits
        ],
        'skipped_resources': skipped_resource_ids,
    }

    if top_resources:
        # include the top resources if requested
        response['top_resources'] = [
            {trim_index_name(bucket['key']): bucket['doc_count']}
            for bucket in result.aggs.to_dict()['indexes']['buckets']
        ]
    timer.add_event('response')

    # allow plugins to modify the fields object
    for plugin in PluginImplementations(IVersionedDatastore):
        response = plugin.datastore_multisearch_modify_response(response)
    timer.add_event('response_modifiers')

    log_query(query, 'multisearch')
    timer.add_event('log')

    if timings:
        response['timings'] = timer.to_dict()
    return response


@action(schema.datastore_create_slug(), help.datastore_create_slug)
def datastore_create_slug(
    context,
    query=None,
    query_version=None,
    version=None,
    resource_ids=None,
    resource_ids_and_versions=None,
    pretty_slug=True,
    nav_slug=False,
):
    """
    Creates a slug for the given multisearch parameters and returns it. This slug can be
    used, along with the resolve_slug action, to retrieve, at any point after the slug
    is created, the query parameters passed to this action. The slug is unique for the
    given query parameters and passing the same query parameters again at a later point
    will result in the same slug being returned.

    :param context: the context dict from the action call
    :param query: the query dict. If None (default) then an empty query is used
    :param query_version: the version of the query schema the query is using. If None (default) then
                          the latest query schema version is used
    :param version: the version to search the data at. If None (default) the current time is used
    :param resource_ids: the list of resource to search. If None (default) then all the resources
                         the user has access to are queried. If a list of resources are passed then
                         any resources not accessible to the user will be removed before querying
    :param resource_ids_and_versions: a dict of resources and versions to search each of them at.
                                      This allows precise searching of each resource at a specific
                                      parameter. If None (default) then the resource_ids parameter
                                      is used together with the version parameter. If this parameter
                                      is provided though, it takes priority over the resource_ids
                                      and version parameters.
    :param pretty_slug: whether to produce a "pretty" slug or not. If True (the default) a selection
                        of 2 adjectives and an animal will be used to create the slug, otherwise if
                        False, a uuid will be used
    :param nav_slug: if this is True, a temporary navigational slug will be produced
                     instead of a standard slug
    :return: a dict containing the slug and whether it was created during this function call or not
    """
    if query is None:
        query = {}
    if query_version and query_version.lower().startswith('v0'):
        # this is an old/basic query so we need to convert it first
        query = convert_to_multisearch(query)
        query_version = None
    if query_version is None:
        query_version = get_latest_query_version()

    try:
        if nav_slug:
            is_new, slug = create_nav_slug(
                context, query, version, resource_ids, resource_ids_and_versions
            )
        else:
            is_new, slug = create_slug(
                context,
                query,
                query_version,
                version,
                resource_ids,
                resource_ids_and_versions,
                pretty_slug=pretty_slug,
            )
    except (jsonschema.ValidationError, InvalidQuerySchemaVersionError) as e:
        raise toolkit.ValidationError(e.message)

    if slug is None:
        raise toolkit.ValidationError('Failed to generate new slug')

    return {
        'slug': slug.get_slug_string(),
        'is_new': is_new,
        'is_reserved': False
        if nav_slug
        else slug.reserved_pretty_slug == slug.get_slug_string(),
    }


@action(
    schema.datastore_resolve_slug(),
    help.datastore_resolve_slug,
    toolkit.side_effect_free,
)
def datastore_resolve_slug(slug):
    """
    Resolves the given slug and returns the query parameters used to create it.

    :param slug: the slug
    :return: the query parameters and the creation time in a dict
    """
    # try resolving the slug first
    resolved = resolve_slug(slug)
    if resolved:
        result = {
            k: getattr(resolved, k)
            for k in (
                'query',
                'query_version',
                'version',
                'resource_ids',
                'resource_ids_and_versions',
            )
        }
        result['created'] = resolved.created.isoformat()
        if result.get('query_version') == 'v0':
            result['query'] = convert_to_multisearch(result['query'])
            result['query_version'] = get_latest_query_version()
        return result

    # then check if it's a query DOI
    if plugin_loaded('query_dois'):
        from ckanext.query_dois.model import QueryDOI
        from ckan import model

        resolved = model.Session.query(QueryDOI).filter(QueryDOI.doi == slug).first()
        if resolved:
            if resolved.query_version == 'v0':
                query = convert_to_multisearch(resolved.query)
                query_version = get_latest_query_version()
            else:
                query = resolved.query
                query_version = resolved.query_version
            return {
                'query': query,
                'query_version': query_version,
                'version': resolved.requested_version,
                'resource_ids': list(resolved.resources_and_versions.keys()),
                'resource_ids_and_versions': resolved.resources_and_versions,
                'created': resolved.timestamp.isoformat(),
            }

    # if both slug and DOI have failed
    raise toolkit.ValidationError('Slug not found')


@action(
    schema.datastore_field_autocomplete(),
    help.datastore_field_autocomplete,
    toolkit.side_effect_free,
)
def datastore_field_autocomplete(context, text='', resource_ids=None, lowercase=False):
    """
    Given a text value, finds fields that contain the given text from the given resource
    (or all resource if no resources are passed).

    :param context: the context dict from the action call
    :param text: the text to search with (default is an empty string)
    :param resource_ids: a list of resources to find fields from, if None (the default) all resource
                         fields are searched
    :param lowercase: whether to do a lowercase check or not, essentially whether to be case
                      insensitive. Default: True, be case insensitive.
    :return: the fields and the resources they came from as a dict
    """
    # figure out which resources should be searched
    resource_ids = get_available_datastore_resources(context, resource_ids)
    if not resource_ids:
        raise toolkit.ValidationError(
            "The requested resources aren't accessible to this user"
        )

    mappings = get_mappings(resource_ids)

    fields = defaultdict(dict)

    for index, mapping in mappings.items():
        resource_id = unprefix_index(index)

        for field_path, config in iter_data_fields(mapping):
            if any(
                text in (part.lower() if lowercase else part) for part in field_path
            ):
                fields['.'.join(field_path)][resource_id] = {
                    'type': config['type'],
                    'fields': {
                        f: c['type'] for f, c in config.get('fields', {}).items()
                    },
                }

    return {
        'count': len(fields),
        'fields': fields,
    }


@action(
    schema.datastore_guess_fields(),
    help.datastore_guess_fields,
    toolkit.side_effect_free,
)
def datastore_guess_fields(
    context,
    query=None,
    query_version=None,
    version=None,
    resource_ids=None,
    resource_ids_and_versions=None,
    size=10,
    ignore_groups=None,
):
    """
    Guesses the fields that are most relevant to show with the given query.

    If only one resource is included in the search then the requested number of fields from the
    resource at the required version are returned in ingest order if the details are available.

    If multiple resources are queried, the most common fields across the resource under search are
    returned. The fields are grouped together in an attempt to match the same field name in
    different cases across different resources. The most common {size} groups are returned.

    The groups returned are ordered firstly by the number of resources they appear in in descending
    order, then if there are ties, the number of records the group finds is used and this again is
    ordered in a descending fashion.

    :param context: the context dict from the action call
    :param query: the query
    :param query_version: the query schema version
    :param version: the version to search at
    :param resource_ids: the resource ids to search in
    :param resource_ids_and_versions: a dict of resource ids -> versions to search at
    :param size: the number of groups to return
    :param ignore_groups: a list of groups to ignore from the results (default: None)
    :return: a list of groups
    """
    # provide some more complex defaults for some parameters if necessary
    if query is None:
        query = {}
    if query_version is None:
        query_version = get_latest_query_version()
    ignore_groups = (
        set(g.lower() for g in ignore_groups) if ignore_groups is not None else set()
    )

    try:
        # validate and translate the query into an elasticsearch-dsl Search object
        validate_query(query, query_version)
        search = translate_query(query, query_version)
    except (jsonschema.ValidationError, InvalidQuerySchemaVersionError) as e:
        raise toolkit.ValidationError(e.message)

    # figure out which resources we're searching
    resource_ids, skipped_resource_ids = determine_resources_to_search(
        context, resource_ids, resource_ids_and_versions
    )
    if not resource_ids:
        raise toolkit.ValidationError(
            "The requested resources aren't accessible to this user"
        )

    if version is None:
        version = to_timestamp(datetime.now())
    # add the version filter necessary given the parameters and the resources we're searching
    version_filter = determine_version_filter(
        version, resource_ids, resource_ids_and_versions
    )
    search = search.filter(version_filter)

    # add the size parameter, we don't want any records back
    search = search.extra(size=0)

    resource_ids = find_searched_resources(search, resource_ids)

    all_fields = get_all_fields(resource_ids)
    for group in ignore_groups:
        all_fields.ignore(group)

    # allow plugins to modify the fields object
    for plugin in PluginImplementations(IVersionedDatastore):
        all_fields = plugin.datastore_modify_guess_fields(resource_ids, all_fields)

    if len(resource_ids) == 1:
        resource_id = resource_ids[0]
        if resource_ids_and_versions is None:
            up_to_version = version
        else:
            up_to_version = resource_ids_and_versions[resource_id]
        return get_single_resource_fields(
            all_fields, resource_id, up_to_version, search, size
        )
    else:
        size = max(0, min(size, 25))
        return select_fields(all_fields, search, size)


@action(
    schema.datastore_value_autocomplete(),
    help.datastore_value_autocomplete,
    toolkit.side_effect_free,
)
def datastore_value_autocomplete(
    context,
    field,
    prefix,
    query=None,
    query_version=None,
    version=None,
    resource_ids=None,
    resource_ids_and_versions=None,
    size=20,
    after=None,
):
    """
    Returns a list of values in alphabetical order from the given field that start with
    the given prefix. The values have to be from the provided resource ids and be from
    documents that match the given query. The after parameter can be used to get the
    next set of values providing pagination. The resulting list is limited to a maximum
    size of 20 values.

    :param context: the context dict from the action call
    :param field: the field to get the values from
    :param prefix: the prefix value to search with (this can be missing/blank to just return the
                   first values)
    :param query: the query
    :param query_version: the query schema version
    :param version: the version to search at
    :param resource_ids: the resource ids to search in
    :param resource_ids_and_versions: a dict of resource ids -> versions to search at
    :param size: the number of values to return (max 500)
    :param after: the after value to use (provides pagination)
    :return: a list of values that match the given prefix
    """
    # provide some more complex defaults for some parameters if necessary
    if query is None:
        query = {}
    if query_version is None:
        query_version = get_latest_query_version()
    # limit the size so that it is between 1 and 500
    size = max(1, min(size, 500))

    try:
        # validate and translate the query into an elasticsearch-dsl Search object
        validate_query(query, query_version)
        search = translate_query(query, query_version)
    except (jsonschema.ValidationError, InvalidQuerySchemaVersionError) as e:
        raise toolkit.ValidationError(e.message)

    # figure out which resources we're searching
    resource_ids, skipped_resource_ids = determine_resources_to_search(
        context, resource_ids, resource_ids_and_versions
    )
    if not resource_ids:
        raise toolkit.ValidationError(
            "The requested resources aren't accessible to this user"
        )

    # add the version filter necessary given the parameters and the resources we're searching
    version_filter = determine_version_filter(
        version, resource_ids, resource_ids_and_versions
    )
    search = search.filter(version_filter)

    # only add the prefix filter to the search if one is provided
    if prefix:
        search = search.filter('prefix', **{prefix_field(field): prefix})

    # we don't need any results so set the size to 0
    search = search.extra(size=0)

    # modify the search so that it has the aggregation required to get the autocompletion results
    search.aggs.bucket(
        'field_values',
        'composite',
        size=size,
        sources={field: A('terms', field=prefix_field(field), order='asc')},
    )
    # if there's an after included, add it into the aggregation
    if after:
        search.aggs['field_values'].after = {field: after}

    # add the resource indexes we're searching on
    search = search.index(
        [prefix_resource(resource_id) for resource_id in resource_ids]
    )

    # create a multisearch for this one query - this ensures there aren't any issues with the length
    # of the URL as the index list is passed as a part of the body
    multisearch = MultiSearch(using=common.ES_CLIENT).add(search)

    # run the search and get the only result from the search results list
    result = next(iter(multisearch.execute()))

    # get the results we're interested in
    agg_result = result.aggs.to_dict()['field_values']
    # return a dict of results, but only include the after details if there are any to include
    response = {
        'values': [bucket['key'][field] for bucket in agg_result['buckets']],
    }
    if 'after_key' in agg_result:
        response['after'] = agg_result['after_key'][field]
    if skipped_resource_ids:
        result['skipped_resources'] = skipped_resource_ids
    return response


@action(
    schema.datastore_hash_query(), help.datastore_hash_query, toolkit.side_effect_free
)
def datastore_hash_query(query=None, query_version=None):
    """
    Hashes the given query at the given query schema and returns the hex digest.

    :param query: the query dict
    :param query_version: the query version
    :return: the hex digest of the query
    """
    if query is None:
        query = {}
    if query_version is None:
        query_version = get_latest_query_version()

    try:
        validate_query(query, query_version)
    except (jsonschema.ValidationError, InvalidQuerySchemaVersionError) as e:
        raise toolkit.ValidationError(e.message)

    return hash_query(query, query_version)


@action(
    schema.datastore_edit_slug(), help.datastore_edit_slug, toolkit.side_effect_free
)
def datastore_edit_slug(context, current_slug, new_reserved_slug):
    slug = resolve_slug(current_slug)
    if slug is None:
        raise toolkit.Invalid(f'The slug {current_slug} does not exist')
    if slug.reserved_pretty_slug and not context['auth_user_obj'].sysadmin:
        raise toolkit.NotAuthorized(
            'Only sysadmins can replace existing reserved slugs.'
        )
    slug.reserved_pretty_slug = new_reserved_slug.lower()
    slug.commit()
    return slug.as_dict()


@action(
    schema.datastore_multisearch_counts(),
    help.datastore_multisearch_counts,
    toolkit.side_effect_free,
)
def datastore_multisearch_counts(
    context,
    query=None,
    query_version=None,
    version=None,
    resource_ids=None,
    resource_ids_and_versions=None,
) -> Dict[str, int]:
    """
    Efficiently counts the number of records in each of the given resources matching the
    given query. A dict of resource IDs -> count is returned. If no records in a
    resource match the query then it will appear in the dict with a count value of 0.

    :param context: the context dict from the action call
    :param query: the query dict. If None (default) then an empty query is used
    :param query_version: the version of the query schema the query is using. If None
                          (default) then the latest query schema version is used
    :param version: the version to search the data at. If None (default) the current
                    time is used
    :param resource_ids: the list of resource to search. If None (default) then all the
                         resources the user has access to are queried. If a list of
                         resources are passed then any resources not accessible to the
                         user will be removed before querying
    :param resource_ids_and_versions: a dict of resources and versions to search each of
                                      them at. This allows precise searching of each
                                      resource at a specific parameter. If None
                                      (default) then the resource_ids parameter is used
                                      together with the version parameter. If this
                                      parameter is provided though, it takes priority
                                      over the resource_ids and version parameters.
    :return: a dict of resource IDs -> count
    """
    # provide some more complex defaults for some parameters if necessary
    if query is None:
        query = {}
    if query_version is None:
        query_version = get_latest_query_version()

    try:
        # validate and translate the query into an elasticsearch-dsl Search object
        validate_query(query, query_version)
        search = translate_query(query, query_version)
    except (jsonschema.ValidationError, InvalidQuerySchemaVersionError) as e:
        raise toolkit.ValidationError(e.message)

    # figure out which resources we're searching
    resource_ids, skipped_resource_ids = determine_resources_to_search(
        context, resource_ids, resource_ids_and_versions
    )
    if not resource_ids:
        raise toolkit.ValidationError(
            "The requested resources aren't accessible to this user"
        )

    # add the version filter necessary given the parameters and the resources we're
    # searching
    version_filter = determine_version_filter(
        version, resource_ids, resource_ids_and_versions
    )
    search = search.filter(version_filter)

    # add the resource indexes we're searching on
    search = search.index(
        [prefix_resource(resource_id) for resource_id in resource_ids]
    )
    # no results please, we aren't going to use them
    search = search.extra(size=0)
    # use an aggregation to get the hit count of each resource, set the size to the
    # number of resources we're querying to ensure we get all counts in one go and don't
    # have to paginate with a composite agg
    search.aggs.bucket("counts", "terms", field="_index", size=len(resource_ids))

    # create a multisearch for this one query - this ensures there aren't any issues
    # with the length of the URL as the index list is passed as a part of the body
    multisearch = MultiSearch(using=common.ES_CLIENT).add(search)

    # run the search and get the only result from the search results list
    result = next(iter(multisearch.execute()))

    # build the response JSON
    counts = {
        trim_index_name(bucket["key"]): bucket["doc_count"]
        for bucket in result.aggs.to_dict()["counts"]["buckets"]
    }
    # add resources that didn't have any hits into the counts dict too
    counts.update(
        {resource_id: 0 for resource_id in resource_ids if resource_id not in counts}
    )
    return counts



import json
import re

from ckan.plugins import toolkit
from ckanext.datastore.logic.schema import json_validator, unicode_or_json_validator
from ckantools.validators import list_validator, list_of_strings, object_validator
from .arg_objects import QueryArgs, DerivativeArgs, ServerArgs, NotifierArgs

# grab all the validator functions upfront
boolean_validator = toolkit.get_validator('boolean_validator')
ignore_missing = toolkit.get_validator('ignore_missing')
int_validator = toolkit.get_validator('int_validator')
not_missing = toolkit.get_validator('not_missing')
not_empty = toolkit.get_validator('not_empty')
resource_id_exists = toolkit.get_validator('resource_id_exists')
email_validator = toolkit.get_validator('email_validator')


def url_safe(value, context):
    """
    Checks if the value is safe to be included in a URL as a slug.

    :param value: the value to check
    :param context: the context in which to check
    """
    if not re.match('^[A-Za-z0-9-_]+$', value):
        raise toolkit.Invalid(
            'Only a-z, 0-9, hyphens (-) and underscores (_) are valid characters'
        )
    else:
        return value


def datastore_search():
    return {
        'resource_id': [not_empty, str, resource_id_exists],
        'q': [ignore_missing, unicode_or_json_validator],
        'filters': [ignore_missing, json_validator],
        'limit': [ignore_missing, int_validator],
        'offset': [ignore_missing, int_validator],
        'fields': [ignore_missing, list_of_strings()],
        'sort': [ignore_missing, list_of_strings()],
        # add an optional version (if it's left out we default to current)
        'version': [ignore_missing, int_validator],
        # if a facets list is included then the top 10 most frequent values for each of the fields
        # listed will be returned along with estimated counts
        'facets': [ignore_missing, list_of_strings()],
        # the facet limits dict allows precise control over how many top values to return for each
        # facet in the facets list
        'facet_limits': [ignore_missing, json_validator],
        'run_query': [ignore_missing, boolean_validator],
        'after': [ignore_missing, json_validator],
    }


def datastore_create():
    return {
        'resource_id': [ignore_missing, str, resource_id_exists],
    }


def datastore_upsert():
    return {
        'resource_id': [not_empty, str, resource_id_exists],
        'replace': [not_missing, boolean_validator],
        'version': [ignore_missing, int_validator],
        # 'records': [ignore_missing, list_of_dicts_validator],
    }


def datastore_delete():
    return {
        'resource_id': [ignore_missing, str, resource_id_exists],
        'version': [ignore_missing, int_validator],
    }


def datastore_get_record_versions():
    return {
        'resource_id': [not_empty, str, resource_id_exists],
        'id': [not_empty, int],
    }


def datastore_autocomplete():
    return {
        'resource_id': [not_empty, str, resource_id_exists],
        'q': [ignore_missing, unicode_or_json_validator],
        'filters': [ignore_missing, json_validator],
        'limit': [ignore_missing, int_validator],
        'after': [ignore_missing, str],
        'field': [not_empty, str],
        'term': [not_missing, str],
        # add an optional version (if it's left out we default to current)
        'version': [ignore_missing, int_validator],
    }


def datastore_reindex():
    return {
        'resource_id': [not_empty, str, resource_id_exists],
    }


def datastore_get_rounded_version():
    return {
        'resource_id': [not_empty, str, resource_id_exists],
        'version': [ignore_missing, int_validator],
    }


def datastore_search_raw():
    return {
        'resource_id': [not_empty, str, resource_id_exists],
        'search': [ignore_missing, json_validator],
        'version': [ignore_missing, int_validator],
        'raw_result': [ignore_missing, boolean_validator],
        'include_version': [ignore_missing, boolean_validator],
    }


def datastore_ensure_privacy():
    return {
        'resource_id': [ignore_missing, str, resource_id_exists],
    }


def datastore_multisearch():
    return {
        'query': [ignore_missing, json_validator],
        'version': [ignore_missing, int_validator],
        'query_version': [ignore_missing, str],
        'resource_ids': [ignore_missing, list_of_strings()],
        'after': [ignore_missing, list_validator],
        'size': [ignore_missing, int_validator],
        'top_resources': [ignore_missing, boolean_validator],
        'resource_ids_and_versions': [ignore_missing, json_validator],
        'timings': [ignore_missing, boolean_validator],
    }


def datastore_field_autocomplete():
    return {
        'text': [ignore_missing, str],
        'resource_ids': [ignore_missing, list_of_strings()],
        'lowercase': [ignore_missing, boolean_validator],
    }


def datastore_value_autocomplete():
    return {
        'field': [not_empty, str],
        'prefix': [ignore_missing, str],
        'query': [ignore_missing, json_validator],
        'version': [ignore_missing, int_validator],
        'query_version': [ignore_missing, str],
        'resource_ids': [ignore_missing, list_of_strings()],
        'after': [ignore_missing, str],
        'size': [ignore_missing, int_validator],
        'resource_ids_and_versions': [ignore_missing, json_validator],
    }


def datastore_create_slug():
    return {
        'query': [ignore_missing, json_validator],
        'version': [ignore_missing, int_validator],
        'query_version': [ignore_missing, str],
        'resource_ids': [ignore_missing, list_of_strings()],
        'resource_ids_and_versions': [ignore_missing, json_validator],
        'pretty_slug': [ignore_missing, boolean_validator],
        'nav_slug': [ignore_missing, boolean_validator],
    }


def datastore_resolve_slug():
    return {
        'slug': [str],
    }


def datastore_count():
    return {
        'resource_ids': [ignore_missing, list_of_strings()],
        'version': [ignore_missing, int_validator],
    }


def datastore_queue_download():
    return {
        'query': [not_missing, object_validator(QueryArgs)],
        'file': [
            not_missing,
            object_validator(DerivativeArgs),
        ],  # called 'file' instead of derivative to make its purpose clearer to the end user
        'server': [ignore_missing, object_validator(ServerArgs)],
        'notifier': [ignore_missing, object_validator(NotifierArgs)],
    }


def datastore_regenerate_download():
    return {
        'download_id': [not_missing, str],
        'server': [ignore_missing, object_validator(ServerArgs)],
        'notifier': [ignore_missing, object_validator(NotifierArgs)],
    }


def datastore_guess_fields():
    return {
        'query': [ignore_missing, json_validator],
        'query_version': [ignore_missing, str],
        'version': [ignore_missing, int_validator],
        'resource_ids': [ignore_missing, list_of_strings()],
        'resource_ids_and_versions': [ignore_missing, json_validator],
        'size': [ignore_missing, int_validator],
        'ignore_groups': [ignore_missing, list_of_strings()],
    }


def datastore_hash_query():
    return {
        'query': [ignore_missing, json_validator],
        'query_version': [ignore_missing, str],
    }


def datastore_is_datastore_resource():
    return {'resource_id': [not_missing, not_empty, resource_id_exists]}


def datastore_edit_slug():
    return {
        'current_slug': [str, not_missing, not_empty],
        'new_reserved_slug': [str, not_missing, not_empty, url_safe],
    }


def datastore_multisearch_counts():
    return {
        'query': [ignore_missing, json_validator],
        'version': [ignore_missing, int_validator],
        'query_version': [ignore_missing, str],
        'resource_ids': [ignore_missing, list_of_strings()],
        'resource_ids_and_versions': [ignore_missing, json_validator],
    }



import re

from ckan.lib import mailer
from ckan.plugins import toolkit
from .base import BaseNotifier

default_html_body = '''
<html lang="en">
<body>
{0}
</body>
</html>
'''.strip()


class EmailNotifier(BaseNotifier):
    name = 'email'

    def __init__(self, request, emails, **kwargs):
        self.email_addresses = emails if isinstance(emails, list) else [emails]
        super(EmailNotifier, self).__init__(request, **kwargs)

    def notify_start(self):
        content, content_html = self.start_text()
        body = content.strip()
        body_html = default_html_body.format(content_html)

        for address in self.email_addresses:
            mailer.mail_recipient(
                recipient_email=address,
                recipient_name='Downloader',
                subject='Data download started',
                body=body,
                body_html=body_html,
            )

    def notify_end(self, file_url):
        content, content_html = self.end_text(file_url)
        body = content.strip()
        body_html = default_html_body.format(content_html)

        for address in self.email_addresses:
            mailer.mail_recipient(
                recipient_email=address,
                recipient_name='Downloader',
                subject='Data download complete',
                body=body,
                body_html=body_html,
            )

    def notify_error(self):
        content, content_html = self.error_text()
        body = content.strip()
        body_html = default_html_body.format(content_html)

        for address in self.email_addresses:
            mailer.mail_recipient(
                recipient_email=address,
                recipient_name='Downloader',
                subject='Data download failed',
                body=body,
                body_html=body_html,
            )

    @classmethod
    def validate_args(cls, type_args):
        emails = type_args.get('emails')
        if not emails:
            raise toolkit.Invalid('Email address must be provided.')
        emails = emails if isinstance(emails, list) else [emails]
        email_rgx = re.compile(r'(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)')
        for e in emails:
            if not email_rgx.fullmatch(e):
                raise toolkit.Invalid('Email address appears to be invalid.')
        return True



import itertools
from collections import Counter, defaultdict

import re
from elasticsearch_dsl import MultiSearch, Q
from elasticsearch_dsl.query import Bool

from .utils import chunk_iterator
from .. import common
from ..datastore_utils import (
    prefix_resource,
    iter_data_fields,
    unprefix_index,
    prefix_field,
)
from ..importing.details import get_all_details


def get_mappings(resource_ids, chunk_size=5):
    """
    Return a dict of mappings for the given resource ids.

    :param resource_ids: a list of resource ids
    :param chunk_size: the number of mappings to get at a time
    :return: a dict of mappings
    """
    mappings = {}
    for chunk in chunk_iterator(map(prefix_resource, resource_ids), chunk_size):
        mappings.update(common.ES_CLIENT.indices.get_mapping(','.join(chunk)))
    return mappings


class Fields(object):
    """
    Class representing the fields from a set of resources.

    Fields are grouped with other fields that have the same name (case-insensitive
    match) allowing the discovery of fields that are common across multiple resources.
    Note that nested fields only match if the whole case-insensitive path matches.
    """

    def __init__(self, skip_ids=True):
        '''
        :param skip_ids: whether to skip the _id columns or not (at any nesting level)
        '''
        self.skip_ids = skip_ids

        # this counts the number of resources a group appears in
        self.group_counts = Counter()
        # this stores the group names, the fields in each group and the resources each field is
        # included in - i.e. group -> {field -> resource ids}
        self.groups = {}
        self.ignore_groups = set()
        self.force_groups = []

    def add(self, field_path, resource_id):
        """
        Adds the given field path from the given resource id into the object's stores.

        :param field_path: a tuple representing a field's path, for root level fields this should be
                           a tuple with a single element like ('genus', ) whereas for nested fields
                           2 or more elements will be present, like ('associatedMedia', 'category').
                           The field path is joined using a '.' to create a full path for each
                           field.
        :param resource_id: the resource id the field belongs to
        """
        if self.skip_ids and field_path[-1] == '_id':
            return

        field = '.'.join(field_path)
        group = field.lower()

        self.group_counts[group] += 1

        if group not in self.groups:
            self.groups[group] = defaultdict(list)
        self.groups[group][field].append(resource_id)

    def ignore(self, group):
        if hasattr(group, 'match') and callable(group.match):
            self.ignore_groups.add(group)
        else:
            self.ignore_groups.add(re.compile(group, re.IGNORECASE))

    def force(self, group):
        if group not in self.force_groups:
            self.force_groups.append(group.lower())

    def get_forced_groups(self):
        for group in self.force_groups:
            if group in self.group_counts:
                yield group, self.group_counts[group], self.groups.get(group, {})

    def top_groups(self):
        """
        Generator which yields the groups with the highest resource representation from
        highest to lowest.

        :return: a generator which yields a 3-tuple on each iteration. The 3-tuple contains the
                 group name, the resource count and a dict of containing all the field names in the
                 group and the resources they appear in (field name -> list of resource ids)
        """
        # return a sorted list in reverse count order, secondarily sorted by group name ascending
        # h/t https://stackoverflow.com/a/23033745. We sort by alphabetical secondarily to ensure
        # there is a stable order of these groups
        for group, count in sorted(
            self.group_counts.most_common(), key=lambda x: (-x[1], x[0])
        ):
            # if the group is in the force_groups list then skip it
            if group in self.force_groups:
                continue

            # ignore groups that have been specifically ignored
            if any(ignore.match(group) for ignore in self.ignore_groups):
                continue

            yield group, count, self.groups.get(group, {})

    def get_searches(self, search):
        groups = itertools.chain(self.get_forced_groups(), self.top_groups())
        for group_name, resource_count, variants in groups:
            shoulds = []
            indexes = []
            for variant, resources_in_group in variants.items():
                shoulds.append(Q('exists', field=prefix_field(variant)))
                indexes.extend(
                    prefix_resource(resource_id) for resource_id in resources_in_group
                )

            # yield the group tuple and an elasticsearch-dsl object for the group's fields
            yield (group_name, resource_count, variants), search.index(indexes).filter(
                Bool(should=shoulds, minimum_should_match=1)
            )

    def is_forced(self, group):
        return group in self.force_groups


def get_all_fields(resource_ids):
    """
    Retrieve the fields available across all of the given resources. The Fields object
    returned contains the fields from all versions of the resources.

    :param resource_ids: the resource ids to get the fields from
    :return: a Fields object
    """
    index_names = [prefix_resource(resource_id) for resource_id in resource_ids]
    mappings = get_mappings(resource_ids)

    fields = Fields()
    for index_name in index_names:
        resource_id = unprefix_index(index_name)
        for field_path, config in iter_data_fields(mappings[index_name]):
            # this catches a corner case where an object type has been specifically defined for a
            # field but no values have been added to the field. Normally, when values exist for a
            # field, we only get the nested field names and not the root name back from
            # iter_data_fields but if no values have been set the object type field will not have
            # any properties and it itself is returned
            if config['type'] == 'object':
                continue
            fields.add(field_path, resource_id)

    return fields


def select_fields(all_fields, search, number_of_groups):
    """
    Selects the fields from the given Fields object which are most common across the
    given resource ids. The search parameter is used to limit the records that
    contribute fields to the returned selection. The fields returned must appear in the
    search in at least one resource with at least one value present.

    :param all_fields: a Fields object
    :param search: an elasticsearch-dsl search object
    :param number_of_groups: the number of groups to select from the Fields object and return
    :return: a list of groups, each group is a dict containing:
                - "group" - the group name
                - "count" - the number of resources its fields appear in
                - "records" - the number of records the group's fields appear in
                - "fields" - the fields that make up the group along with the resource ids they come
                             from
                - "forced" - whether the field was forced into being included, or whether it was
                             included organically
    """
    selected_fields = []
    # make sure we don't get any hits back, we're only interested in the counts
    search = search.extra(size=0)

    # iterate over the groups and searches in chunks
    for chunk in chunk_iterator(
        all_fields.get_searches(search), chunk_size=number_of_groups
    ):
        groups, searches = zip(*chunk)
        # create a multisearch for all the searches in the group
        multisearch = MultiSearch(using=common.ES_CLIENT)
        for search in searches:
            multisearch = multisearch.add(search)

        for (group, count, fields), response in zip(groups, multisearch.execute()):
            if all_fields.is_forced(group) or response.hits.total > 0:
                # a field from this group has values in the search result, add it to the selection
                selected_fields.append(
                    dict(
                        group=group,
                        count=count,
                        records=response.hits.total,
                        fields=fields,
                        forced=all_fields.is_forced(group),
                    )
                )

        if len(selected_fields) >= number_of_groups:
            break

    def group_sorter(the_group):
        # this sorts the groups ensuring forced groups are first, in the order they were forced,
        # then the groups with highest count and then the ones with the highest number of records
        if the_group['forced']:
            # use 0 0 to ensure that the base order of the groups is maintained for forced groups
            return True, 0, 0
        else:
            return False, the_group['count'], the_group['records']

    # sort the returned selected list by count and secondly records
    return sorted(selected_fields, key=group_sorter, reverse=True)


def get_single_resource_fields(
    all_fields, resource_id, version, search, number_of_groups
):
    """
    Retrieves the fields for a single given resource. The fields are returned in the
    same format as the select_fields function above.

    :param all_fields: a Fields object
    :param resource_id: the resource id to be searched in
    :param version: the version we're searching at
    :param search: an elasticsearch-dsl search object
    :param number_of_groups: the maximum number of groups to return
    :return: a list of groups, each group is a dict containing:
            - "group" - the group name
            - "count" - the number of resources its fields appear in (will always be 1)
            - "records" - the number of records the group's fields appear in
            - "fields" - the fields that make up the group along with the resource ids they come
                         from (the list of resource ids will always just be the one under search)
    """
    # retrieve the datastore resource details if there are any
    all_details = get_all_details(resource_id, up_to_version=version)
    if all_details:
        # the all_details variable is an OrderedDict in ascending version order. We want to iterate
        # in descending version order though so that we respect the column order at the version
        # we're at before respecting any data from previous versions
        for details in reversed(all_details.values()):
            for field_name in details.get_columns():
                all_fields.force(field_name)

    return select_fields(all_fields, search, number_of_groups)



records = [
    {
        'scientificName': 'Boops boops',
        'img': 'img-url-here',
        'urlSlug': 'boops-boops',
        'emptyField': 'not actually empty',
        'group': 'a',
    },
    {
        'scientificName': 'Felis catus',
        'img': 'cat-picture-here',
        'urlSlug': 'felis-catus',
        'group': 'b',
    },
]

records_addtl = [
    {
        'scientificName': 'Gigantiops destructor',
        'img': 'add-picture-here',
        'urlSlug': 'g-destructor',
        'colour': 'green',
        'group': 'b',
    },
    {
        'scientificName': 'Tyrannosaurus rex',
        'img': 'img-url-here',
        'urlSlug': 't-rex',
        'colour': 'blue',
        'group': 'b',
    },
]
